from gempy_legacy.core.model import Project
import qgrid


class QgridModelIntegration(object):
    """
    Class that handles the changes done interactively in qgrid and updates a Model object.

    """
    def __init__(self, geo_model: Project, plot_object=None):
        # TODO add on all to update from data_object and plots?

        self._geo_model = geo_model
        self._plot_object = plot_object
        if plot_object is not None:
            pass
            # Check if the window is already open
            # self.set_vtk_object(plot_object)

            # if hasattr(plot_object, 'interactor'):
            #     self._plot_object.set_surface_points()
            #     self._plot_object.set_orientations()

        self.qgrid_fo = self.set_interactive_df('surfaces')
        self.qgrid_se = self.set_interactive_df('series')
        self.qgrid_fa = self.set_interactive_df('faults')
        self.qgrid_fr = self.set_interactive_df('faults_relations')
        self.qgrid_in = self.set_interactive_df('surface_points')
        self.qgrid_or = self.set_interactive_df('orientations')
        self.qgrid_op = self.set_interactive_df('options')
        self.qgrid_kr = self.set_interactive_df('kriging')
        self.qgrid_re = self.set_interactive_df('rescale')

    def update_plot(self):
        if self._plot_object is not None:
            self._plot_object.update_surfaces()
        else:
            pass

    def set_interactive_df(self, data_type: str):
        if data_type == 'surfaces':
            self.qgrid_fo = self.create_surfaces_qgrid()
            return self.qgrid_fo
        elif data_type == 'series':
            self.qgrid_se = self.create_series_qgrid()
            return self.qgrid_se
        elif data_type == 'faults':
            self.qgrid_fa = self.create_faults_qgrid()
            return self.qgrid_fa
        elif data_type == 'faults_relations':
            self.qgrid_fr = self.create_faults_relations_qgrid()
            return self.qgrid_fr
        elif data_type == 'surface_points':
            self.qgrid_in = self.create_surface_points_qgrid()
            return self.qgrid_in
        elif data_type == 'orientations':
            self.qgrid_or = self.create_orientations_qgrid()
            return self.qgrid_or
        elif data_type == 'options':
            self.qgrid_op = self.create_options_qgrid()
            return self.qgrid_op
        elif data_type == 'kriging':
            self.qgrid_kr = self.create_kriging_parameters_qgrid()
            return self.qgrid_kr
        elif data_type == 'rescale':
            self.qgrid_re = self.create_rescaling_data_qgrid()
            return self.qgrid_re

        else:
            raise AttributeError('data_type must be either surfaces, series, faults, faults_relations,'
                                 ' surface_points, orientations,'
                                 ' options, kriging or rescale. UPDATE message')

    def set_vtk_object(self, vtk_object):
        self._plot_object = vtk_object
        if not hasattr(vtk_object, 'interactor'):
            self._plot_object.set_surface_points()
            self._plot_object.set_orientations()

    def update_qgrd_objects(self):
        qgrid_objects = [self.qgrid_fo, self.qgrid_se, self.qgrid_fa,
                         self.qgrid_in,
                         self.qgrid_or, self.qgrid_op, self.qgrid_kr, self.qgrid_re]

        for e, qgrid_object in enumerate(qgrid_objects):
           # print('qgrid_object' + str(e) )
            if e == 1:
                qgrid_object.df = self._geo_model._stack.df.reset_index().rename(columns={'index': 'series_names'}).astype(
                    {'series_names': str})

            self.update_qgrid_object(qgrid_object)

    @staticmethod
    def update_qgrid_object(qgrid_object):
        qgrid_object._ignore_df_changed = True
        # make a copy of the user's dataframe
        qgrid_object._df = qgrid_object.df.copy()

        # insert a column which we'll use later to map edits from
        # a filtered version of this df back to the unfiltered version
        qgrid_object._df.insert(0, qgrid_object._index_col_name, range(0, len(qgrid_object._df)))

        # keep an unfiltered version to serve as the starting point
        # for filters, and the state we return to when filters are removed
        qgrid_object._unfiltered_df = qgrid_object._df.copy()

        qgrid_object._update_table(update_columns=True, fire_data_change_event=True)
        qgrid_object._ignore_df_changed = False

    def get(self, data_type: str):

        if data_type is 'surfaces':
            self.qgrid_fo._update_df()
            return self.qgrid_fo
        elif data_type is 'series':
            self.qgrid_se._update_df()
            return self.qgrid_se
        elif data_type is 'faults':
            self.qgrid_fa._update_df()
            return self.qgrid_fa
        elif data_type is 'faults_relations':
            self.qgrid_fr._update_df()
            return self.qgrid_fr
        elif data_type is 'surface_points':
            self.qgrid_in._update_df()
            return self.qgrid_in
        elif data_type is 'orientations':
            self.qgrid_or._update_df()
            return self.qgrid_or
        elif data_type is 'options':
            self.qgrid_op._update_df()
            return self.qgrid_op
        elif data_type is 'kriging':
            self.qgrid_kr._update_df()
            return self.qgrid_kr
        elif data_type is 'rescale':
            self.qgrid_re._update_df()
            return self.qgrid_re

        else:
            raise AttributeError('data_type must be either surfaces, series, faults, surface_points, orientations,'
                                 'options, kriging or rescale. UPDATE message')

    def create_faults_qgrid(self):
        faults_object = self._geo_model._faults

        qgrid_widget = qgrid.show_grid(faults_object.df,
                                       show_toolbar=False,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True},
                                       column_options={'editable': True},
                                       column_definitions={'isFinite': {'editable': True}})

        def handle_set_is_fault(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)

            if event['column'] == 'isFault':
                idx = event['index']
                #      cat_idx = qgrid_widget.df.loc[idx, 'series_names']

                self._geo_model.set_is_fault([idx], toggle=True)
                self.update_plot()

            if event['column'] == 'isFinite':
                idx = event['index']
                #      cat_idx = qgrid_widget.df.loc[idx, 'series_names']

                self._geo_model.set_is_finite_fault([idx], toggle=True)
                self.update_plot()

            self.update_qgrd_objects()
            self.qgrid_fr._rebuild_widget()

        qgrid_widget.on('cell_edited', handle_set_is_fault)

        return qgrid_widget

    def create_rescaling_data_qgrid(self):
        rescaling_object = self._geo_model._rescaling

        qgrid_widget = qgrid.show_grid(rescaling_object.df,
                                       show_toolbar=False,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True})

        def handle_row_edit(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
                print(qgrid_widget._df)

            value = event['new']
            self._geo_model.modify_rescaling_parameters(event['column'], value)
            self.update_qgrd_objects()
            self.update_plot()

        qgrid_widget.on('cell_edited', handle_row_edit)
        return qgrid_widget

    def create_options_qgrid(self):
        options_object = self._geo_model._additional_data.options

        qgrid_widget = qgrid.show_grid(options_object.df,
                                       show_toolbar=False,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True})

        def handle_row_edit(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
                print(self)
                print(qgrid_widget._df)

            if event['column'] == 'verbosity':
                import numpy as np
                value = np.fromstring(event['new'][1:-1], sep=',')
            else:
                value = event['new']
            self._geo_model.modify_options(event['column'], value)
            self.update_qgrd_objects()
            self.update_plot()

        qgrid_widget.on('cell_edited', handle_row_edit)
        return qgrid_widget

    def create_kriging_parameters_qgrid(self):
        kriging_parameters_object = self._geo_model._additional_data.kriging_data

        qgrid_widget = qgrid.show_grid(kriging_parameters_object.df,
                                       show_toolbar=False,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True},
                                       )

        def handle_row_edit(event, widget, debug=False):

            if debug is True:
                print(event)
                print(widget)
                print(qgrid_widget._df)

            self._geo_model.modify_kriging_parameters(event['column'], event['new'])
            self.update_qgrd_objects()
            self.update_plot()

        qgrid_widget.on('cell_edited', handle_row_edit)
        return qgrid_widget

    def create_faults_relations_qgrid(self):
        faults_object = self._geo_model._faults

        # We need to add the qgrid special columns to categories if does not exist
        try:
            faults_object.faults_relations_df.columns = faults_object.faults_relations_df.columns.add_categories(
                ['index', 'qgrid_unfiltered_index'])
        except ValueError:
            pass

        qgrid_widget = qgrid.show_grid(faults_object.faults_relations_df,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True},
                                       show_toolbar=False)

        def handle_set_fault_relation(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)

            # This data frame is quite independent to anything else:
            self._geo_model.set_fault_relation(qgrid_widget.get_changed_df().values)

            # faults_object.faults_relations_df.update(qgrid_widget.get_changed_df())

            self.update_qgrd_objects()
            self.update_plot()
        qgrid_widget.on('cell_edited', handle_set_fault_relation)
        return qgrid_widget

    def create_surface_points_qgrid(self):
        surface_points_object = self._geo_model._surface_points
        self._geo_model.set_default_surfaces()
        self._geo_model.set_default_surface_point(plot_object=self._plot_object)

        qgrid_widget = qgrid.show_grid(
            surface_points_object.df,
            show_toolbar=True,
            grid_options={'sortable': False, 'highlightSelectedCell': True},
            column_options={'editable': False},
            column_definitions={'X': {'editable': True},
                                'Y': {'editable': True},
                                'Z': {'editable': True},
                                'surface': {'editable': True}})

        def handle_row_surface_points_add(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
                print(qgrid_widget._df)
            idx = event['index']

            xyzs = qgrid_widget._df.loc[idx, ['X', 'Y', 'Z', 'surface']]
            self._geo_model.add_surface_points(*xyzs, idx=int(idx), plot_object=self._plot_object)
            self.update_qgrd_objects()

        def handle_row_surface_points_delete(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            idx = event['indices']

            self._geo_model.delete_surface_points(idx, plot_object=self._plot_object)
            self.update_qgrd_objects()

        def handle_cell_surface_points_edit(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)

            column = event['column']
            idx = event['index']
            value = event['new']

            try:
                self._geo_model.modify_surface_points(idx, **{column: value, 'plot_object': self._plot_object})
            except AssertionError:
                pass
            self.update_qgrd_objects()

        qgrid_widget.on('row_removed', handle_row_surface_points_delete)
        qgrid_widget.on('row_added', handle_row_surface_points_add)
        qgrid_widget.on('cell_edited', handle_cell_surface_points_edit)
        return qgrid_widget

    def create_orientations_qgrid(self):
        orientations_object = self._geo_model._orientations
        self._geo_model.set_default_orientation(plot_object=self._plot_object)

        qgrid_widget = qgrid.show_grid(
            orientations_object.df,
            show_toolbar=True,
            grid_options={'sortable': False, 'highlightSelectedCell': True},
            column_options={'editable': False},
            column_definitions={'X': {'editable': True},
                                'Y': {'editable': True},
                                'Z': {'editable': True},
                                'G_x': {'editable': True},
                                'G_y': {'editable': True},
                                'G_z': {'editable': True},
                                'azimuth': {'editable': True},
                                'dip': {'editable': True},
                                'pole': {'editable': True},
                                'surface': {'editable': True}})

        def handle_row_orientations_add(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
                print(qgrid_widget._df)
            idx = event['index']
            xyzs = qgrid_widget._df.loc[idx, ['X', 'Y', 'Z', 'surface']]
            gxyz = qgrid_widget._df.loc[idx, ['G_x', 'G_y', 'G_z']]
            self._geo_model.add_orientations(*xyzs, pole_vector=gxyz.values, idx=int(idx),
                                             plot_object=self._plot_object)
            self.update_qgrd_objects()

        def handle_row_orientations_delete(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            idx = event['indices']

            self._geo_model.delete_orientations(idx, plot_object=self._plot_object)
            self.update_qgrd_objects()

        def handle_cell_orientations_edit(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)

            column = event['column']
            idx = event['index']
            value = event['new']

            self._geo_model.modify_orientations(idx, **{column: value, 'plot_object': self._plot_object})
            self.update_qgrd_objects()

        qgrid_widget.on('row_removed', handle_row_orientations_delete)
        qgrid_widget.on('row_added', handle_row_orientations_add)
        qgrid_widget.on('cell_edited', handle_cell_orientations_edit)
        return qgrid_widget

    def create_surfaces_qgrid(self):
        surface_object = self._geo_model._surfaces

        surface_object.set_default_surface_name()
        qgrid_widget = qgrid.show_grid(surface_object.df, show_toolbar=True,
                                       grid_options={'sortable': False, 'highlightSelectedCell': True},
                                       column_options={'editable': True},
                                       column_definitions={'id': {'editable': False},
                                                           'isFault': {'editable': False},
                                                           'isActive': {'editable': False},
                                                           'vertices': {'editable': False},
                                                           'edges': {'editable': False},
                                                           'sfai': {'editable': False},
                                                           'isBasement': {'editable': False}})

        def handle_row_surface_add(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            idx = event['index']
            self._geo_model.add_surfaces(['surface' + str(idx+1)])
            #surface_object.add_surface(['surface' + str(idx)])
            self.update_qgrd_objects()

            self.qgrid_in._rebuild_widget()
            self.qgrid_or._rebuild_widget()

        def handle_row_surface_delete(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            idx = event['indices']
            self._geo_model.delete_surfaces(idx)
            #surface_object.delete_surface(idx)
            self.update_qgrd_objects()
            self.qgrid_in._rebuild_widget()
            self.qgrid_or._rebuild_widget()

        def handle_cell_surface_edit(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            if event['column'] == 'surface':
                self._geo_model.rename_surfaces({event['old']: event['new']})
            if event['column'] == 'series':
                idx = event['index']
                new_series = event['new']
                self._geo_model.map_stack_to_surfaces({new_series: surface_object.df.loc[idx, ['surface']]},
                                                      set_series=False, sort_geometric_data=True)

            if event['column'] == 'order_surfaces':
                idx = event['index']
                try:
                    self._geo_model.modify_order_surfaces(int(event['new']), idx)
                except AssertionError:
                    pass

            self.update_qgrd_objects()

            if event['column'] == 'surface':
                self.qgrid_in._rebuild_widget()
                self.qgrid_or._rebuild_widget()

        qgrid_widget.on('row_removed', handle_row_surface_delete)
        qgrid_widget.on('row_added', handle_row_surface_add)
        qgrid_widget.on('cell_edited', handle_cell_surface_edit)

        return qgrid_widget

    def create_series_qgrid(self):
        series_object = self._geo_model._stack

        # I need to do a serious hack because qgrid does not accept categorical index yet
        qgrid_widget = qgrid.show_grid(
            series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype({'series_names': str}),
            show_toolbar=True,
            grid_options={'sortable': False, 'highlightSelectedCell': True},
            column_options={'editable': True},
            column_definitions={'order_series': {'editable': True}, 'isActive': {'editable': False},
                                })

        def handle_row_series_add(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
                print(series_object.df.reset_index())

            idx = event['index']
            self._geo_model.add_features(['series' + str(idx)])

            # This is because qgrid does not accept editing indeces. We enable the modification of the series name
            # by reindexing the df and change another column
            #self._geo_model.faults.faults_relations_df.columns = self._geo_model.faults.faults_relations_df.columns.add_categories(
            #    ['index', 'qgrid_unfiltered_index'])

            qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
                {'series_names': str})
            self.update_qgrd_objects()
            self.qgrid_fo._rebuild_widget()
            self.qgrid_fr._rebuild_widget()

        def handle_row_series_delete(event, widget, debug=False):
            if debug is True:
                print(event)
                print(widget)
            idx = event['indices']
            cat_idx = qgrid_widget.df.loc[idx, 'series_names']

            self._geo_model.delete_features(cat_idx)

            qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
                {'series_names': str})
            self.update_qgrd_objects()
            self.qgrid_fo._rebuild_widget()
            self.qgrid_fr._rebuild_widget()

        def handle_cell_series_edit(event, widget, debug=False):
            idx = event['index']
            cat_idx = qgrid_widget.df.loc[idx, 'series_names']
            if debug is True:
                print(event)
                print(widget)
                print(cat_idx)
                print(series_object.df.index)
            if event['column'] == 'series_names':
                self._geo_model.rename_features({event['old']: event['new']})
            if event['column'] == 'BottomRelation':
                #series_object.df.loc[cat_idx, 'BottomRelation'] = event['new']
                self._geo_model.set_bottom_relation(cat_idx, event['new'])
            if event['column'] == 'order_series':
                idx = event['index']
                try:
                    self._geo_model.modify_order_features(int(event['new']), idx)
                except AssertionError:
                    pass

            self.update_plot()

         #   self._geo_model.update_from_series(rename_series={event['old']: event['new']})
            # Hack for the faults relations
            print(self._geo_model._faults.faults_relations_df.columns)
           # self._geo_model.faults.faults_relations_df.columns = self._geo_model.faults.faults_relations_df.columns.add_categories(
          #      ['index', 'qgrid_unfiltered_index'])

            qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
                {'series_names': str})

            self.update_qgrd_objects()
            self.qgrid_fo._rebuild_widget()
            self.qgrid_fr._rebuild_widget()

        qgrid_widget.on('row_removed', handle_row_series_delete)
        qgrid_widget.on('row_added', handle_row_series_add)
        qgrid_widget.on('cell_edited', handle_cell_series_edit)

        return qgrid_widget




"""
Simulation of the Perth Basin model with PFLOTRAN
=================================================

"""

# %% 
#
# The purpose of this example is to introduce how to export GemPy geological 
# model to the finite volume reactive transport simulator `PFLOTRAN <https://www.pflotran.org>`_.

# %% 
# The Perth Basin model
# ^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
# First, create the perf bassin model as in `examples/examples/real/Perth_bassin.py`.

import gempy_legacy as gp
data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
geo_model = gp.create_model('Perth_Basin')
gp.init_data(geo_model,
             extent=[337000, 400000, 6640000, 6710000, -18000, 1000],
             resolution=[100, 100, 100],
             path_i=data_path + "/data/input_data/Perth_basin/Paper_GU2F_sc_faults_topo_Points.csv",
             path_o=data_path + "/data/input_data/Perth_basin/Paper_GU2F_sc_faults_topo_Foliations.csv")
del_surfaces = ['Cadda', 'Woodada_Kockatea', 'Cattamarra']
geo_model.delete_surfaces(del_surfaces, remove_data=True)
ret = gp.map_stack_to_surfaces(geo_model,
                          {"fault_Abrolhos_Transfer": ["Abrolhos_Transfer"],
                           "fault_Coomallo": ["Coomallo"],
                           "fault_Eneabba_South": ["Eneabba_South"],
                           "fault_Hypo_fault_W": ["Hypo_fault_W"],
                           "fault_Hypo_fault_E": ["Hypo_fault_E"],
                           "fault_Urella_North": ["Urella_North"],
                           "fault_Urella_South": ["Urella_South"],
                           "fault_Darling": ["Darling"],
                           "Sedimentary_Series": ['Cretaceous',
                                                  'Yarragadee',
                                                  'Eneabba',
                                                  'Lesueur',
                                                  'Permian']
                           })
order_series = ["fault_Abrolhos_Transfer",
                "fault_Coomallo",
                "fault_Eneabba_South",
                "fault_Hypo_fault_W",
                "fault_Hypo_fault_E",
                "fault_Urella_North",
                "fault_Darling",
                "fault_Urella_South",
                "Sedimentary_Series", 'Basement']
geo_model.reorder_series(order_series)

geo_model.surface_points.df.dropna(inplace=True)
geo_model.orientations.df.dropna(inplace=True)
geo_model.set_is_fault(["fault_Abrolhos_Transfer",
                        "fault_Coomallo",
                        "fault_Eneabba_South",
                        "fault_Hypo_fault_W",
                        "fault_Hypo_fault_E",
                        "fault_Urella_North",
                        "fault_Darling",
                        "fault_Urella_South"])
                        

fr = geo_model.faults.faults_relations_df.values
fr[:, :-2] = False
ret = geo_model.set_fault_relation(fr)

geo_model.set_topography(source='random')
interp_data = gp.set_interpolator(geo_model,
                                  compile_aesara=True,
                                  aesara_optimizer='fast_run', gradient=False,
                                  dtype='float32')
gp.compute_model(geo_model)

# %%
# Here is the results:

gp.plot_3d(geo_model, show_topography=True)


# %% 
#
# Export the model in PFLOTRAN format
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Two format are available for PFLOTRAN and are automatically determined by the extension:
# 
# * `ascii`: every file are writed into human readable files (`ugi` extension).
# * `hdf5`: all the mesh information are stored into one binary HDF5 file.
# 
# In this example, the `ascii` format is chosen, so the model is saved
# under the name ``perth_bassin_mesh.ugi``:
# 

# sphinx_gallery_thumbnail_path = '_static/permeability_pflotran.png'
import gempy_legacy.utils.export as export
export.export_pflotran_input(geo_model, path='', filename="perth_basin_mesh.ugi")


# %% 
#
#PFLOTRAN ASCII format description
#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
#When using the `ascii` file format, several file are created. Mesh 
#informations are stored into the main file ``perth_bassin_mesh.ugi``. Also, 
#the stratigraphic information are stored in separed files, with the
#name of the formation and the extension `.vs` 
#(for Volume Submesh). For example, the Perth Basin model got several
#formations which are saved in PFLOTRAN format under the following files:
#
#* basement -> basement.vs
#* Cretaceous -> Cretaceous.vs
#* Eneabba -> Eneabba.vs
#* Lesueur -> Lesueur.vs
#* and so on..
#

# %% 
#Group based on stratigraphic information are loaded in PFLOTRAN with the
#card ``REGION`` (see `here <https://www.pflotran.org/documentation/user_guide/cards/subsurface/region_card.html>`_).
#For example for the `Cretaceous` formation:
# .. code-block:: python
#  
#  REGION Cretaceous
#    FILE ./Cretaceous.vs
#  END
#
#
#A sample PFLOTRAN input file is provided to correctly read the GemPy 
#output ``pflotran_perth_bassin.in``.
#It basically read the model and perform a saturated flow simulation 
#using different hydraulic conductivity for each formation.
#
#  .. figure:: permeability_pflotran.png
#    :width: 600
#
#Note: GemPy export a file named ``topography_surface.ss`` that represent
#horizontal faces defining the topography. They can be used to apply a 
#boundary condition such as specifying a rain. They are imported in PFLOTRAN by:
# .. code-block:: python
#  
#  REGION topo
#    FILE ./topography_surface.ss
#  END

# %% 
# 
#PFLOTRAN HDF5 format description
#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
#At the contrary, if the `hdf5` format is used (extension ``.h5``), GemPy
#only create one binary file. The stratigraphic information are stored
#in the HDF5 file, and can also be assessed in PFLOTRAN using the ``REGION`` 
#card. For example, for the `Cretaceous` formation and the topographic
#surface:
#
# .. code-block:: python
#   
#  REGION Cretaceous
#    FILE ./perth_bassin_mesh.h5
#  END
#  REGION Topography_surface
#    FILE ./perth_bassin_mesh.h5
#  END
#
#

# %%
#
#Getting help with PFLOTRAN
#^^^^^^^^^^^^^^^^^^^^^^^^^^
#
#For issue relating with PFLOTRAN, please see the discussion group `here <https://groups.google.com/g/pflotran-users>`_.
#
#
#




from typing import Iterable
import gempy_legacy as gp
import pandas as pd
import numpy as np


def loop2gempy(
        contacts_file: str,
        orientations_file: str,
        bbox: Iterable,
        groups_file: str,
        model_base: float,
        model_top: float, dtm_reproj_file: str = None,
        faults_contact: str = None,
        faults_orientations: str = None,
        faults_faults_rel: str = None,
        faults_groups_rel: str = None,
        faults_rel_matrix = None,
        model_name: str = None,
        compute: bool = True,
        vtk: bool = False,
        vtk_path: str = None,
        image_2d: bool = False,
        plot_3d_kwargs=None
):
    """ Calculate the model using gempy as backend.

        At the moment there is not support for finite faults since gempy does not
         accept passing the ellipsoid parameters directly.

        :param contacts_file (str): path of contacts file
        :param orientations_file: path of orientations file
        :param bbox: model bounding box
        :param groups_file: path of groups file
        :param model_base: z value of base of model
        :param model_top: z value of top of model
        :param dtm_reproj_file: path of dtm file
        :param faults_contact: path of contacts file with fault data
        :param faults_orientations: path of orientations file with fault data
        :param faults_rel_matrix: bool matrix describing the interaction between groups. Rows offset columns
        :param faults_groups_rel: bool matrix describing the interaction between faults and features
        :param faults_faults_rel: bool matrix describing the interaction between faults and faults
        :param model_name: name of the model
        :param compute (bool): Default True. Whether or not compute the model
        :param vtk (bool): Default False. Whether or not visualize the model
        :param vtk_path (str): Default None. Path of vtk output directory
        :param plot_3d_kwargs (dict): kwargs for `gempy.plot_3d`
        :return: gempy.Project
    """
    if plot_3d_kwargs is None:
        plot_3d_kwargs = {}

    contacts = []
    orientations = []

    contacts.append(
        pd.read_csv(
            contacts_file,
            sep=',',
            names=['X', 'Y', 'Z', 'formation'],
            header=1
        )
    )

    if faults_contact is not None:
        contacts.append(
            pd.read_csv(
                faults_contact,
                sep=',',
                names=['X', 'Y', 'Z', 'formation'],
                header=1
            )
        )

    orientations.append(
        pd.read_csv(
            orientations_file,
            sep=',',
            names=['X', 'Y', 'Z', 'azimuth', 'dip', 'polarity', 'formation'],
            header=1
        )
    )

    if faults_orientations is not None:
        orientations.append(
            pd.read_csv(
                faults_orientations,
                sep=',',
                names=['X', 'Y', 'Z', 'azimuth', 'dip', 'polarity', 'formation'],
                header=1
            )
        )

    if faults_faults_rel is not None and faults_groups_rel is not None:
        ff_ = pd.read_csv(faults_faults_rel).set_index('fault_id')
        fg_ = pd.read_csv(faults_groups_rel).set_index('group')
        p_ = pd.concat((ff_, fg_), axis=0, sort=True)
        faults_rel_matrix = pd.concat((p_, fg_.T), axis=1, sort=True).fillna(0).values

    surface_points_ready = pd.concat(contacts, sort=True)
    surface_points_ready.reset_index(inplace=True, drop=False)
    orientation_ready = pd.concat(orientations, sort=True)
    orientation_ready.reset_index(inplace=True, drop=False)

    if model_name is None:
        model_name = 'loop2gempy'

    geo_model = gp.create_model(model_name)
    gp.init_data(
        geo_model,
        extent=[bbox[0], bbox[2], bbox[1], bbox[3], model_base, model_top],
        resolution=[50, 50, 50],
        orientations_df=orientation_ready,
        surface_points_df=surface_points_ready
    )

    # Load Topology
    if dtm_reproj_file is not None:
        if type(dtm_reproj_file) is str:
            source = 'gdal'
            topo_kwarg = {'filepath': dtm_reproj_file}
        elif type(dtm_reproj_file) is np.ndarray:
            source = 'numpy'
            topo_kwarg = {'array': dtm_reproj_file}
        else:
            raise AttributeError('dtm_proj_file must be either a path to gdal or a'
                                 'numpy array with values')
        geo_model.set_topography(source=source, **topo_kwarg)

    # Stack Processing
    contents = np.genfromtxt(groups_file, delimiter=',', dtype='U100')[1:, 4:-1]

    map_series_to_surfaces = {}
    for pair in contents:
        map_series_to_surfaces.setdefault(pair[1], []).append(pair[0])

    gp.map_stack_to_surfaces(geo_model, map_series_to_surfaces,
                             remove_unused_series=False)

    order_formations = geo_model.stack.df.index.drop('Default series')

    # Get the unassigned series as faults
    if faults_contact is not None and faults_orientations is not None:
        faults_pair = geo_model._surfaces.df.groupby('series').get_group('Default series')[[
            'surface']].values[:, 0]
        faults_pair_dict = dict(zip(faults_pair, faults_pair))
        gp.map_stack_to_surfaces(geo_model, faults_pair_dict, remove_unused_series=True)
        # Grabbing the order of faults and Formations
        ordered_features = np.append(faults_pair, order_formations)
        # Sorting series
        geo_model.reorder_features(ordered_features)
        geo_model.set_is_fault(faults_pair)

        # Faults relation
        geo_model.set_fault_relation(faults_rel_matrix)

    geo_model.add_surfaces('basement')

    # Check if there is features without data and delete it
    try:
        f_ = geo_model.surfaces.df.groupby('isActive').get_group(False)
        features_without_data = f_['series']
        geo_model.delete_features(features_without_data, remove_surfaces=True, remove_data=True)
    except KeyError:
        pass

    if faults_contact is not None and faults_orientations is not None:
        get_fault_names = geo_model.stack.df.groupby(['isActive', 'isFault']).get_group(
            (True, True)).index
        geo_model._surfaces.colors.make_faults_black(get_fault_names)

    try:
        colours = pd.read_csv(groups_file).set_index('code')['colour']
        # Drop formations that do not exist in surfaces
        colours = colours.loc[colours.index.isin(geo_model.surfaces.df['surface'])].to_dict()

        geo_model._surfaces.colors.change_colors(colours)
    except KeyError:
        pass

    if compute is True:
        gp.set_interpolator(geo_model, dtype='float64',
                            # verbose=['solve_kriging']
                            )

        # Increasing nugget effect
        geo_model.modify_surface_points(
            geo_model.surface_points.df.index,
            smooth=0.1
        )

        geo_model.modify_orientations(
            geo_model.orientations.df.index,
            smooth=0.01
        )

        new_range = geo_model.get_additional_data().loc[('Kriging', 'range'), 'values'] * 0.5
        geo_model.modify_kriging_parameters('range', new_range)

        gp.compute_model(geo_model)

    if vtk is True:
        gp.plot_3d(geo_model, show_topography=True,
                   image=image_2d,
                   show_lith=True,
                   **plot_3d_kwargs
                   )

    if vtk_path is not None:
        gp._plot.export_to_vtk(geo_model, path=vtk_path, name=model_name + '.vtk',
                               voxels=False, block=None, surfaces=True)

    return geo_model


"""
Export a geological model from GemPy to use in MOOSE
====================================================

"""

# %%
import gempy_legacy as gp

# %%
# Creating a geological model
# ---------------------------
# 
# The procedure of generating a geological model is presented in detail in
# `Chapter
# 1-1 <https://nbviewer.jupyter.org/github/cgre-aachen/gempy/blob/master/notebooks/tutorials/ch1-1_Basics.ipynb>`__
# of the GemPy tutorials, so it will only be briefly presented here
# 

# %% 
# Initiate a model
geo_model = gp.create_model('tutorial_moose_exp')
data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'

# %% 
# Import data from CSV-files with setting the resolution and model extent
gp.init_data(geo_model, [0, 2000., 0, 2000., 0, 2000.], [50, 50, 80],
             path_o=data_path + "/data/input_data/tut_chapter1/simple_fault_model_orientations.csv",
             path_i=data_path + "/data/input_data/tut_chapter1/simple_fault_model_points.csv",
             default_values=True)

# %% 
# present the units and series
geo_model.surfaces

# %% 
# combine units in series and make two series, as the fault needs its own
gp.map_stack_to_surfaces(geo_model,
                         {"Fault_Series": 'Main_Fault',
                          "Strat_Series": ('Sandstone_2', 'Siltstone', 'Shale', 'Sandstone_1', 'basement')},
                         remove_unused_series=True)

# set the fault series to be fault object
geo_model.set_is_fault(['Fault_Series'], change_color=False)

# %% 
# check whether series were assigned correctly
geo_model.surfaces

# %%
# Model generation
# ----------------
# 
# After loading in the data, we set it up for interpolation and compute
# the model.
# 

# %% 
# set up interpolator
gp.set_interpolator(geo_model,
                    compile_aesara=True,
                    aesara_optimizer='fast_compile',
                    verbose=[])

# %% 
# compute the model
gp.compute_model(geo_model, compute_mesh=False);

# %% 
# have a look at the data and computed model
gp.plot_3d(geo_model)

# %%
# Exporting the Model to MOOSE
# ----------------------------
# 
# The voxel-model above already is the same as a model discretized in a
# hexahedral grid, so my immediately be used as input in a simulation
# tool, e.g. `MOOSE <https://mooseframework.org/>`__. For this, we need to
# access to the unit IDs assigned to each voxel in GemPy. The array
# containing these IDs is called ``lith_block``.
# 

# %% 
ids = geo_model.solutions.lith_block
print(ids)

# %%
# This array has the shape of ``(x,)`` and would be immediately useful, if
# GemPy and the chosen simulation code would *populate* a grid in the same
# way. Of course, however, that is not the case. This is why we have to
# restructure the ``lith_block`` array, so it can be read correctly by
# MOOSE.
# 

# %% 
# model resolution
nx, ny, nz = geo_model.grid.regular_grid.resolution

# model extent
xmin, xmax, ymin, ymax, zmin, zmax = geo_model.grid.regular_grid.extent

# %%
# These two parameters are important to, a) restructure ``lith_block``,
# and b) write the input file for MOOSE correctly. For a), we need to
# reshape ``lith_block`` again to its three dimensions and *re-flatten* it
# in a *MOOSE-conform* way.
# 

# %% 
# reshape to 3D array
units = ids.reshape((nx, ny, nz))
# flatten MOOSE conform
units = units.flatten('F')

# %%
# | The importance of ``nx, ny, nz`` is apparent from the cell above. But
#   what about ``xmin``, …, ``zmax``?
# | A MOOSE input-file for mesh generation has the following syntax:
# 
# .. code:: python
# 
#    [MeshGenerators]
#      [./gmg]
#        type = GeneratedMeshGenerator
#        dim = 3
#        nx = 50
#        ny = 50
#        nz = 80
#        xmin = 0.0
#        xmax = 2000.0
#        yim = 0.0
#        ymax = 2000.0
#        zmin = 0.0
#        zmax = 2000.0
#        block_id = '1 2 3 4 5 6'
#        block_name = 'Main_Fault Sandstone_2 Siltstone Shale Sandstone_1 basement'
#      [../]
# 
#      [./subdomains]
#        type = ElementSubdomainIDGenerator
#        input = gmg
#        subdomain_ids = ' ' # here you paste the transformed lith_block vector
#      [../]
#    []
# 
#    [Mesh]
#      type = MeshGeneratorMesh
#    []
# 
# So these parameters are required inputs in the ``[MeshGenerators]``
# object in the MOOSE input file. ``GemPy`` has a method to directly
# create such an input file, stored in ``gempy.utils.export.py``.
# 
# The following cell shows how to call the method:
# 

# %%

# sphinx_gallery_thumbnail_path = '_static/GemPy_model_combined.png'
import gempy_legacy.utils.export as export
export.export_moose_input(geo_model, path='')

# %%
# This method automatically stores a file
# ``geo_model_units_moose_input.i`` at the specified path. Either this
# input file could be extended with parameters to directly run a
# simulation, or it is used just for creating a mesh. In the latter case,
# the next step would be, to run the compiled MOOSE executable witch the
# optional flag ``--mesh-only``.
# 
# E.g. with using the `PorousFlow
# module <https://mooseframework.inl.gov/modules/porous_flow/>`__:
# 
# .. code:: bash
# 
#    $path_to_moose/moose/modules/porous_flow/porous_flow-opt -i pct_voxel_mesh.i --mesh-only
# 
# How to compile MOOSE is described in their
# `documentation <https://mooseframework.inl.gov/getting_started/index.html>`__.
# 
# The now generated mesh with the name
# ``geo_model_units_moose_input_in.e`` can be used as input for another
# MOOSE input file, which contains the main simulation parameters. To call
# the file with the grid, the following part has to be added in the MOOSE
# simulation input file:
# 
# .. code:: python
# 
#    [Mesh]
#      file = geo_model_units_moose_input_in.e
#    []
# 
# .. raw:: html
# 
#    <hr>
# 
# The final output of the simulation may also be such an ``.e``, which
# can, for instance, be opened with
# `paraview <https://www.paraview.org/>`__. A simulated temperature field
# (purely conductive) of the created model would look like this:
# 
# .. figure:: https://raw.githubusercontent.com/Japhiolite/a-Moose-and-you/master/imgs/GemPy_model_combined.png
#    :alt: gempy_temperature
#
#



import pytest

import numpy as np
import sys, os

sys.path.append("../..")
import gempy_legacy
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

input_path = os.path.dirname(__file__) + '/../input_data'
update_sol = False
test_values = [45, 150, 2500, 67000, 100000]


class TestNoFaults:
    """
    I am testing all block and potential field values so sol is (n_block+n_pot)
    """

    def test_a(self, interpolator):
        """
        2 Horizontal layers with drift 0
        """
        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 10, 0, 10, -10, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_a/test_a_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_a/test_a_Points.csv")

        geo_data.set_aesara_function(interpolator)

        # Compute model
        sol = gempy_legacy.compute_model(geo_data)

        if update_sol:
            np.save(input_path + '/test_a_sol.npy', sol.lith_block[test_values])

        # Load model
        real_sol = np.load(input_path + '/test_a_sol.npy')

        # Checking that the plots do not rise errors
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_a.png', dpi=100)

        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, show_scalar=True)

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)

    def test_b(self, interpolator):
        """
        Two layers a bit curvy, drift degree 1
        """

        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 10, 0, 10, -10, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_b/test_b_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_b/test_b_Points.csv")

        geo_data.set_aesara_function(interpolator)

        # Compute model
        sol = gempy_legacy.compute_model(geo_data)

        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_b.png', dpi=200)

        if update_sol:
            np.save(input_path + '/test_b_sol.npy', sol.lith_block[test_values])

        # Load model
        real_sol = np.load(input_path + '/test_b_sol.npy')

        # Checking that the plots do not rise errors
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25)

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)

    def test_c(self, interpolator):
        """
        Two layers a bit curvy, drift degree 0
        """

        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 10, 0, 10, -10, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_c/test_c_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_c/test_c_Points.csv")

        geo_data.set_aesara_function(interpolator)

        # Compute model
        sol = gempy_legacy.compute_model(geo_data)

        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_c.png', dpi=200)

        if update_sol:
            np.save(input_path + '/test_c_sol.npy', sol.lith_block[test_values])

        # Load model
        real_sol = np.load(input_path + '/test_c_sol.npy')

        # Checking that the plots do not rise errors
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, show_scalar=True)

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)


class TestFaults:

    def test_d(self, interpolator):
        """
        Two layers 1 fault
        """

        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 10, 0, 10, -10, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_d/test_d_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_d/test_d_Points.csv")

        gempy_legacy.map_stack_to_surfaces(geo_data, {'fault1': 'f1', 'series': ('A', 'B')})

        geo_data.set_is_fault('fault1')

        geo_data.set_aesara_function(interpolator)

        # Compute model
        sol = gempy_legacy.compute_model(geo_data)

        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_d.png', dpi=200)

        if update_sol:
            np.save(input_path + '/test_d_sol.npy', sol.lith_block[test_values])

        # Load model
        real_sol = np.load(input_path + '/test_d_sol.npy')

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)

    def test_e(self, interpolator):
        """
        Two layers a bit curvy, 1 fault
        """
        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 10, 0, 10, -10, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_e/test_e_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_e/test_e_Points.csv")

        gempy_legacy.map_stack_to_surfaces(geo_data, {'fault1': 'f1', 'series': ('A', 'B')})
        geo_data.set_is_fault('fault1')

        geo_data.set_aesara_function(interpolator)

        # Compute model
        sol = gempy_legacy.compute_model(geo_data)

        if update_sol:
            np.save(input_path + '/test_e_sol.npy', sol.lith_block[test_values])

        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_e.png', dpi=200)

        # Load model
        real_sol = np.load(input_path + '/test_e_sol.npy')

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)

    def test_f_sort_surfaces(self, interpolator):
        """
        Two layers a bit curvy, 1 fault. Checked with geomodeller
        """

        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 2000, 0, 2000, -2000, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_f/test_f_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_f/test_f_Points.csv")

        gempy_legacy.map_stack_to_surfaces(geo_data, {'fault1': 'MainFault',
                                               'series'       : ('Reservoir',
                                                          'Seal',
                                                          'SecondaryReservoir',
                                                          'NonReservoirDeep'
                                                          ),
                                                      },
                                           )

        geo_data.set_aesara_function(interpolator)
        geo_data.set_is_fault('fault1')

        # Compute model
        sol = gempy_legacy.compute_model(geo_data, sort_surfaces=True)

        if update_sol:
            np.save(input_path + '/test_f_sol.npy', sol.lith_block[test_values])

        real_sol = np.load(input_path + '/test_f_sol.npy')
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.show()
        plt.savefig(os.path.dirname(__file__) + '/../figs/test_f.png', dpi=200)

        gempy_legacy.compute_model(geo_data)
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.show()

        gempy_legacy.compute_model(geo_data)
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.show()

        # We only compare the block because the absolute pot field I changed it
        np.testing.assert_array_almost_equal(np.round(sol.lith_block[test_values]), real_sol, decimal=0)

        ver, sim = gempy_legacy.get_surfaces(geo_data)
        print(ver, sim)

    def test_compute_model_multiple_ranges(self, interpolator):

        # Importing the data from csv files and settign extent and resolution
        geo_data = gempy_legacy.create_data(extent=[0, 2000, 0, 2000, -2000, 0], resolution=[50, 50, 50],
                                            path_o=input_path + "/GeoModeller/test_f/test_f_Foliations.csv",
                                            path_i=input_path + "/GeoModeller/test_f/test_f_Points.csv")

        gempy_legacy.map_stack_to_surfaces(geo_data, {'fault1': 'MainFault',
                                               'series'       : ('Reservoir',
                                                          'Seal',
                                                          'SecondaryReservoir',
                                                          'NonReservoirDeep'
                                                          ),
                                                      },
                                           )

        geo_data.set_aesara_function(interpolator)
        geo_data.set_is_fault('fault1')
        geo_data.modify_kriging_parameters('range', [3000, 3500, 0])
        geo_data._additional_data.kriging_data.set_default_c_o()
        # Compute model
        sol = gempy_legacy.compute_model(geo_data, sort_surfaces=True)
        gempy_legacy.plot.plot_2d(geo_data, cell_number=25, direction='y', show_data=True)
        plt.show()


def test_simple_model_gempy_engine():
    import numpy
    numpy.set_printoptions(precision=3, linewidth=200)

    g = gempy_legacy.create_data("test_engine", extent=[-4, 4, -4, 4, -4, 4], resolution=[4, 1, 4])
    sp = np.array([[-3, 0, 0],
                   [0, 0, 0],
                   [2, 0, 0.5],
                   [2.5, 0, 1.2],
                   [3, 0, 2],
                   [1, 0, .2],
                   [2.8, 0, 1.5]])

    g.set_default_surfaces()

    for i in sp:
        g.add_surface_points(*i, surface="surface1")

    g.add_orientations(-3, 0, 2, pole_vector=(0, 0, 1), surface="surface1")
    g.add_orientations(2, 0, 3, pole_vector=(-.2, 0, .8), surface="surface1")

    g.modify_orientations([0, 1], smooth=0.000000000001)
    g.modify_surface_points(g._surface_points.df.index, smooth=0.0000000001)

    gempy_legacy.set_interpolator(g, verbose=[
        "n_surface_op_float_sigmoid",
        "scalar_field_iter",
        "compare",
        "sigma"
    ])

    g.modify_kriging_parameters("range", 50)
    # g.modify_kriging_parameters("$C_o$", 5 ** 2 / 14 / 3)
    g.modify_kriging_parameters("drift equations", [0])

    import aesara
    dtype = "float32"

    g._interpolator.aesara_graph.i_reescale.set_value(np.cast[dtype](1.))
    g._interpolator.aesara_graph.gi_reescale.set_value(np.cast[dtype](1.))

    gempy_legacy.compute_model(g)

    print(g.additional_data)
    print(g.solutions.scalar_field_matrix)

    gempy_legacy.plot_2d(g)
    print(g.grid.values)

    print(g.solutions.weights_vector)



"""
Transform 2019: Integrating Striplog and GemPy
==============================================

"""

# %% 
# ! pip install welly striplog

# %% 
# Authors: M. de la Varga, Evan Bianco, Brian Burnham and Dieter Werthmüller
# Importing GemPy
import gempy_legacy as gp

# Importing auxiliary libraries
import numpy as np
import pandas as pn
import matplotlib.pyplot as plt
import os
import welly
from welly import Location, Project
import glob
from striplog import Striplog, Legend, Decor

pn.set_option('precision', 2)

# %%
# Creating striplog object
# -----------------------------
# 

# %% 
# get well header coordinates
well_heads = {'alpha': {'kb_coords': (0, 0, 0)},
              'beta': {'kb_coords': (10, 10, 0)},
              'gamma': {'kb_coords': (12, 0, 0)},
              'epsilon': {'kb_coords': (20, 0, 0)}}

# %% 
# Reading tops file
cwd = os.getcwd()
if 'examples' not in cwd:
    data_path = os.getcwd() + '/examples'
else:
    data_path = cwd + '/..'

print(data_path+'/data/input_data/striplog_integration/*.tops')
topsfiles = glob.glob(data_path+'/data/input_data/striplog_integration/*.tops')
topsfiles

# %% 
# Creating striplog object
my_striplogs = []

for file in topsfiles:
    with open(file) as f:
        text = f.read()
        striplog = Striplog.from_csv(text=text)
        my_striplogs.append(striplog)

striplog_dict = {'alpha': my_striplogs[1],
                 'beta': my_striplogs[2],
                 'gamma': my_striplogs[3],
                 'epsilon': my_striplogs[0]}

striplog_dict['alpha'][0]

# %% 
# Plot striplog
f, a = plt.subplots(ncols=4, sharey=True)

for e, log in enumerate(striplog_dict.items()):
    log[1].plot(ax=a[e], legend=None)
f.tight_layout()
plt.show()

# %%
# Striplog to pandas df of bottoms
rows = []
for wellname in striplog_dict.keys():
    for i, interval in enumerate(striplog_dict[wellname]):
        surface_name = interval.primary.lith
        surface_base = interval.base.middle
        x, y = well_heads[wellname]['kb_coords'][:-1]
        series = 1
        rows.append([x, y, surface_base, surface_name, series, wellname])

column_names = ['X', 'Y', 'Z', 'surface', 'series', 'wellname']
df = pn.DataFrame(rows, columns=column_names)
df

# %%
# GemPy model
# ----------------
# 

# %% 
# Create gempy model object
geo_model = gp.create_model('welly_integration')

extent = [-100, 300, -100, 200, -150, 0]
res = [60, 60, 60]

# Initializting model using the striplog df
gp.init_data(geo_model, extent, res, surface_points_df=df)

# %% 
geo_model.surface_points.df.head()

# %% 
geo_model.surfaces

# %% 
dec_list = []
for e, i in enumerate(striplog_dict['alpha']):
    dec_list.append(Decor({'_colour': geo_model.surfaces.df.loc[e, 'color'],
                           'width': None,
                           'component': i.primary,
                           'hatch': None}))


# %% 
# welly plot with gempy colors
# Create Decor list
dec_list = []
for e, i in enumerate(striplog_dict['alpha']):
    dec_list.append(Decor({'_colour': geo_model.surfaces.df.loc[e, 'color'],
                           'width': None,
                           'component': i.primary,
                           'hatch': None}))

# Create legend
legend = Legend(dec_list)
legend

# %% 
# Plot striplogs:
f, a = plt.subplots(ncols=4, sharey=True)

for e, log in enumerate(striplog_dict.items()):
    log[1].plot(ax=a[e], legend=legend)
f.tight_layout()
plt.show()

# %% 
# Modifying the coordinates to make more sense
geo_model.surface_points.df[['X', 'Y']] = geo_model.surface_points.df[['X', 'Y']] * 10
geo_model.surface_points.df['Z'] *= -1

# %% 
# Delete points of the basement surface since we are intepolating bottoms (that surface wont exit).
geo_model.delete_surface_points_basement()

# %% 
# Adding an arbitrary orientation. Remember gempy need an orientation per series
geo_model.set_default_orientation()
geo_model.modify_orientations(0, X=-500)

# %% 
gp.plot_2d(geo_model)


# %%
gp.set_interpolator(geo_model)

# %% 
gp.compute_model(geo_model)

# %% 

p2d = gp.plot_2d(geo_model, cell_number=[30], show_data=True, show=True)

# %%
gp.plot_3d(geo_model)


# %%
# Pinch out model
# ------------------
# 
# As we can see the 3D model generated above does not honor the forth well
# lets fix it. First lets add an unconformity: between the yellow and
# green layer:
# 

# %% 
geo_model.add_features('Unconformity')

# %%
# Now we set the green layer in the second series
# 

# %% 
geo_model.map_stack_to_surfaces({'Uncomformity': ['brian', 'evan', 'dieter']})
geo_model.add_surfaces('basement')

# %%
# Lastly we need to add a dummy orientation to the new series:
# 

# %% 
geo_model.add_orientations(-500, 0, -100, 'dieter', [0, 0, 1])

# %%
# Now we can compute:
# 

# %% 
gp.compute_model(geo_model)

# %% 
p = gp.plot_2d(geo_model, cell_number=[30], show_data=True)
f, a = plt.subplots(ncols=4, sharey=True)

for e, log in enumerate(striplog_dict.items()):
    log[1].plot(ax=a[e], legend=legend)
f.tight_layout()
plt.show()

# %%
# Getting better but not quite there yet. Since the yellow does not show
# up in the last well the pinch out has to happen somewhere before so lets
# add an artifial point to get that shape:
# 

# %% 
geo_model.add_surface_points(200, 0, -75, 'evan');

# %% 
gp.compute_model(geo_model)
p = gp.plot_2d(geo_model, cell_number=[30], show_data=True)
f, a = plt.subplots(ncols=4, sharey=True)

for e, log in enumerate(striplog_dict.items()):
    log[1].plot(ax=a[e], legend=legend)
f.tight_layout()
plt.show()

# %%
# sphinx_gallery_thumbnail_number = 7
gp.plot_3d(geo_model)

# %%
# gp.save_model(geo_model)



from mpl_toolkits import axes_grid1
import matplotlib.pyplot as plt


def add_colorbar(im=None, axes=None, cs=None, label = None, aspect=30, location="right", pad_fraction=1, **kwargs):
    """
    Add a colorbar to a plot (im).
    Args:
        im:             plt imshow
        label:          label of the colorbar
        axes:
        cs:             Contourset
        aspect:         the higher, the smaller the colorbar is
        pad_fraction:
        **kwargs:

    Returns: A perfect colorbar, no matter the plot.

    """
    if axes is None:
        axes = im.axes
  
    cax = axes.inset_axes([1.04, 0.2, 0.05, 0.6])
    if cs:
        cbar = axes.figure.colorbar(cs, ax=axes, cax=cax, location="right", **kwargs)
    else:
        if im is not None:
            cbar = axes.figure.colorbar(im, cax=cax, location="right",  **kwargs)
    cbar.set_label(label)
    return cbar


import gempy_legacy
import os
import numpy as np
import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
input_path = os.path.dirname(__file__)+'/../input_data'


def test_rescaled_marching_cube(interpolator):
    """
    2 Horizontal layers with drift 0
    """
    # Importing the data from csv files and setting extent and resolution
    geo_data = gempy_legacy.create_data('Simple interpolator', [0, 10, 0, 10, -10, 0], [50, 50, 50],
                                        path_o=input_path + "/GeoModeller/test_a/test_a_Foliations.csv",
                                        path_i=input_path + "/GeoModeller/test_a/test_a_Points.csv")

    geo_data.set_aesara_function(interpolator)

    # Compute model
    sol = gempy_legacy.compute_model(geo_data, compute_mesh_options={'rescale': True})
    print(sol.vertices)

    return geo_data


def test_custom_grid_solution(model_horizontal_two_layers):
    """
    Integration test for a gempy model using a custom grid

    2 Horizontal layers with drift 0

    :param interpolator_islith_nofault:
    :return:
    """
    geo_model = model_horizontal_two_layers
    # add a custom grid
    cg = np.array([[5, 5, -9],
                   [5, 5, -5],
                   [5, 5, -5.1],
                   [5, 5, -5.2],
                   [5, 5, -1]])
    values = geo_model.set_custom_grid(cg)
    assert geo_model._grid.active_grids[1]
    # set the aesara function

    # Compute model
    sol = gempy_legacy.compute_model(geo_model, compute_mesh=False)
    assert sol.custom.shape == (2, 1, 5)


def test_masked_marching_cubes(unconformity_model_topo):
    geo_model = unconformity_model_topo
    gp.plot_2d(geo_model)
    plt.show()

    gp.plot_2d(geo_model, regular_grid=geo_model.solutions.mask_matrix_pad[0],
               kwargs_regular_grid={'cmap': 'viridis', 'norm': None})
    
    print(geo_model.solutions.mask_matrix_pad[0])
    plt.show()

    gp.plot_2d(geo_model, regular_grid=geo_model.solutions.mask_matrix_pad[1],
               kwargs_regular_grid={'cmap': 'viridis', 'norm': None}
               )
    print(geo_model.solutions.mask_matrix_pad[1])
    plt.show()

    gp.plot_2d(geo_model, regular_grid=geo_model.solutions.mask_matrix[0],
               kwargs_regular_grid={'cmap': 'viridis', 'norm': None})
    print(geo_model.solutions.mask_matrix[0])
    plt.show()





""" Getters API

"""
from gempy_legacy import Project, Solution
from typing import Union
import warnings
from gempy_legacy.utils.meta import _setdoc_pro


@_setdoc_pro()
def get_data(model: Project, itype='data', numeric=False, verbosity=0):
    """Method to return the data stored in :class:`pandas.DataFrame` within a
    :class:`gempy.core.model.Project` data object.

    Args:
        model: [s_geo_model]
        itype: [s_itype]
        numeric (bool): if True it only returns numerical properties. This may be useful due to memory issues
        verbosity (int): Number of properties shown

    Returns:
        pandas.DataFrame: Data Object df.

    """
    return model.get_data(itype=itype, numeric=numeric, verbosity=verbosity)


def get_surfaces(model_solution: Union[Project, Solution]):
    """Get vertices and simplices of the surface_points for its vtk visualization and further analysis.

    Args:
       model_solution (:class:`Project` or :class:`Solution`)

    Returns:
        list[numpy.array]: vertices, simpleces
    """
    if isinstance(model_solution, Project):
        return model_solution.solutions.vertices, model_solution.solutions.edges
    elif isinstance(model_solution, Solution):
        return model_solution.vertices, model_solution.edges
    else:
        raise AttributeError


def get_interpolator(model: Project):
    return model._interpolator


def get_th_fn(model: Project):
    """Get the compiled aesara function

    Args:
        model (:class:`model.Project`)

    Returns:
        :class:`aesara.compile.function_module.Function`: Compiled function if C or CUDA which computes the interpolation given the input data
            (XYZ of dips, dip, azimuth, polarity, XYZ ref surface_points, XYZ rest surface_points)
    """
    assert getattr(model._interpolator, 'aesara_function', False) is not None, 'aesara has not been compiled yet'

    return model._interpolator.aesara_function


def get_additional_data(model: Project):
    warnings.warn('get_additional_data will be deprecrated in GemPy 2.3 Use '
                  'get(\'additional_data\') instead.', DeprecationWarning)
    return model.get_additional_data()



"""
Greenstone.
===========
"""

# Importing gempy
import gempy_legacy as gp

# Aux imports
import numpy as np
import matplotlib.pyplot as plt
import os

print(gp.__version__)

# %% 
geo_model = gp.create_model('Greenstone')

# %%

data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'

# Importing the data from csv files and settign extent and resolution
geo_model = gp.init_data(geo_model, [696000, 747000, 6863000, 6930000, -20000, 200], [50, 50, 50],
                         path_o=data_path + "/data/input_data/tut_SandStone/SandStone_Foliations.csv",
                         path_i=data_path + "/data/input_data/tut_SandStone/SandStone_Points.csv")

# %% 
gp.plot_2d(geo_model, direction=['z'])

# %% 
gp.map_stack_to_surfaces(geo_model, {"EarlyGranite_Series": 'EarlyGranite',
                                     "BIF_Series": ('SimpleMafic2', 'SimpleBIF'),
                                     "SimpleMafic_Series": 'SimpleMafic1', 'Basement': 'basement'})

# %% 
geo_model.add_surface_values([2.61, 2.92, 3.1, 2.92, 2.61])

# %% 
gp.set_interpolator(geo_model,
                    compile_aesara=True,
                    aesara_optimizer='fast_compile',
                    verbose=[])

# %% 
gp.compute_model(geo_model, set_solutions=True)

# %% 
gp.plot_2d(geo_model, cell_number=[-1], direction=['z'], show_data=False)

# %% 
gp.plot_2d(geo_model, cell_number=[25], direction='x')

# %% 
geo_model.solutions.values_matrix

# %% 
p2d = gp.plot_2d(geo_model, cell_number=[25], block=geo_model.solutions.values_matrix,
           direction=['y'], show_data=True,
           kwargs_regular_grid={'cmap': 'viridis', 'norm':None})

# %%
# sphinx_gallery_thumbnail_number = 5
gp.plot_3d(geo_model)

# %% 
np.save('greenstone_ver', geo_model.solutions.vertices)
np.save('greenstone_edges', geo_model.solutions.edges)

# %%
# Saving the model
# ~~~~~~~~~~~~~~~~
# 

# %% 
# gp.save_model(geo_model, path=os.pardir + '/data/gempy_models')
gp.save_model(geo_model)


# TODO
# - Check the basement layer is not in in SurfacePoints and Orientations

import pandas as pn
import numpy as np

def check_for_nans(array_like):
    if ~(pn.notnull(np.atleast_1d(array_like))).any():
        array_like = None
    return array_like


def check_for_nans_input_aesara():
    pass


def check_kriging():
    # TODO check range and covariance are not 0
    pass

def check_faults_are_set():
    pass

def check_fault_relations(self):
    # Method to check that only older df offset newer ones?
    #
    # try:
    #     # Check if there is already a categories_df
    #     self.df
    #
    #     try:
    #         if any(self.df.columns != series.columns):
    #             series_fault = self.count_faults()
    #             self.df = pn.DataFrame(index=series.columns, columns=['isFault'])
    #             self.df['isFault'] = self.df.index.isin(series_fault)
    #     except ValueError:
    #         series_fault = self.count_faults()
    #         self.df = pn.DataFrame(index=series.columns, columns=['isFault'])
    #         self.df['isFault'] = self.df.index.isin(series_fault)
    #
    #     if series_fault:
    #         self.df['isFault'] = self.df.index.isin(series_fault)
    #
    # except AttributeError:
    #
    #     if not series_fault:
    #         series_fault = self.count_faults()
    #         self.df = pn.DataFrame(index=series.columns, columns=['isFault'])
    #         self.df['isFault'] = self.df.index.isin(series_fault)

    # self.surface_points.loc[:, 'isFault'] = self.surface_points['series'].isin(self.df.index[self.df['isFault']])
    # self.orientations.loc[:, 'isFault'] = self.orientations['series'].isin(
    #     self.df.index[self.df['isFault']])
    pass


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


Module with classes and methods to visualize structural geology data and potential fields of the regional modelling based on
the potential field method.
Tested on Ubuntu 14

Created on 10.04.2019

@author: Miguel de la Varga
"""

import os
import matplotlib.colors as mcolors
import copy
import pandas as pn
import numpy as np
import sys
import gempy_legacy as gp
import warnings


warnings.filterwarnings("ignore",
                        message='.*Conversion of the second argument of issubdtype *.',
                        append=True)
try:
    import vtk
    from vtk.util.numpy_support import numpy_to_vtk
    VTK_IMPORT = True
except ImportError:
    VTK_IMPORT = False

try:
    import steno3d
    STENO_IMPORT = True
except ImportError:
    STENO_IMPORT = False

try:
    import ipyvolume as ipv
    IPV_IMPORT = True
except ImportError:
    IPV_IMPORT = False

import pyvista as pv

class vtkVisualization(object):
    """
    Class to visualize data and results in 3D. Init will create all the render properties while the method render
    model will lunch the window. Using set_surface_points, set_orientations and set_surfaces in between can be chosen what
    will be displayed.

    Args:
        geo_data(gempy.InputData): All values of a DataManagement object
        ren_name (str): Name of the renderer window
        verbose (int): Verbosity for certain functions

    Attributes:
        renWin(vtk.vtkRenderWindow())
        camera_list (list): list of cameras for the distinct renderers
        ren_list (list): list containing the vtk renderers
    """

    def __init__(self, geo_data, ren_name='GemPy 3D-Editor', verbose=0, real_time=False, bg_color=None, ve=1,
                 **kwargs):
        if VTK_IMPORT is False:
            raise ImportError('vtk is not installed. No vtk capabilities are possible')

        self.ve = ve

        self.real_time = real_time
        self.geo_model = geo_data
        self.interp_data = None
        self.layer_visualization = True
        # self.C_LOT = #dict(zip(self.geo_model.surfaces.df['id'], self.geo_model.surfaces.df['color']))
        #
        # for surf, color in self.C_LOT.items(): #convert hex to rgb
        #     self.C_LOT[surf] = mcolors.hex2color(color)

        self.ren_name = ren_name
        # Number of renders
        self.n_ren = 4
        self.id = geo_data._surface_points.df['id'].unique()
        self.surface_name = geo_data._surface_points.df['surface'].unique()

        # Extents
        self.extent = copy.copy(self.geo_model._grid.regular_grid.extent)
        self.extent[-1] = ve * self.extent[-1]
        self.extent[-2] = ve * self.extent[-2]
        _e = self.extent
        self._e_dx = _e[1] - _e[0]
        self._e_dy = _e[3] - _e[2]
        self._e_dz = _e[5] - _e[4]
        self._e_d_avrg = (self._e_dx + self._e_dy + self._e_dz) / 3

        # Create containers of the vtk objectes
        self.s_rend_1 = pn.DataFrame(columns=['val'])
        self.s_rend_2 = pn.DataFrame(columns=['val'])
        self.s_rend_3 = pn.DataFrame(columns=['val'])
        self.s_rend_4 = pn.DataFrame(columns=['val'])

        self.o_rend_1 = pn.DataFrame(columns=['val'])
        self.o_rend_2 = pn.DataFrame(columns=['val'])
        self.o_rend_3 = pn.DataFrame(columns=['val'])
        self.o_rend_4 = pn.DataFrame(columns=['val'])

        # Resolution
        self.res = self.geo_model._grid.regular_grid.resolution

        # create render window, settings
        self.renwin = vtk.vtkRenderWindow()
        self.renwin.SetWindowName(ren_name)

        # Set 4 renderers. ie 3D, X,Y,Z projections
        self.ren_list = self.create_ren_list()

        # create interactor and set interactor style, assign render window
        self.interactor = vtk.vtkRenderWindowInteractor()
        self.interactor.SetRenderWindow(self.renwin)
        self.interactor.AddObserver("KeyPressEvent", self.key_callbacks)

        # 3d model camera for the 4 renders
        self.camera_list = self._create_cameras(self.extent, verbose=verbose)
        # Setting the camera and the background color to the renders
        self.set_camera_backcolor(color=bg_color)

        # Creating the axis
        for e, r in enumerate(self.ren_list):
            # add axes actor to all renderers
            axe = self._create_axes(self.camera_list[e])

            r.AddActor(axe)
            r.ResetCamera()
        # TODO panel does not deal with text
        self.set_text()

    #

    # def restart(self):
    #     try:
    #         self.close_window()
    #     except AttributeError:
    #         pass
    #
    #     self.__init__(self.geo_model, )

    def render_model(self, **kwargs):
        """
        Method to launch the window

        Args:
            size (tuple): Resolution of the window
            fullscreen (bool): Launch window in full screen or not
        Returns:

        """
        # initialize and start the app
        if 'size' not in kwargs:
            kwargs['size'] = (1920, 1080)

        if 'fullscreen' in kwargs:
            self.renwin.FullScreenOn()

        silent = kwargs.get('silent', False)

        self.renwin.SetSize(kwargs['size'])

        if silent is False:
            self.interactor.Initialize()
            self.interactor.Start()

    def set_text(self):
        txt = vtk.vtkTextActor()
        txt.SetInput("Press L to toggle layers visibility \n"
                     "Press R to toggle real time updates \n"
                     "Press H or P to go back to Python \n"
                     "Press Q to quit")
        txtprop = txt.GetTextProperty()
        txtprop.SetFontFamilyToArial()
        txtprop.SetFontSize(18)
        txtprop.SetColor(1, 1, 1)
        txt.SetDisplayPosition(20, 60)

        # assign actor to the renderer
        self.ren_list[0].AddActor(txt)

    def close_window(self):
        # close_window(interactor)
        render_window = self.interactor.GetRenderWindow()
        render_window.Finalize()
        self.interactor.TerminateApp()
        del self.renwin, self.interactor

    def key_callbacks(self, obj, event):
        key = self.interactor.GetKeySym()

        if key == 'h' or key == 'p':
            print('holding... Use vtk.resume to go back to the interactive window')
            self.interactor.ExitCallback()
            self.interactor.holding = True

        if key == 'l':
            if self.layer_visualization is True:
                for layer in self.surf_rend_1:
                    layer.VisibilityOff()
                self.layer_visualization = False
                self.interactor.Render()
            elif self.layer_visualization is False:
                for layer in self.surf_rend_1:
                    layer.VisibilityOn()
                self.layer_visualization = True
                self.interactor.Render()

        if key == 't':
            if self.topo_visualization is True:
                self.topography_surface.VisibilityOff()
                self.topo_visualization = False
                self.interactor.Render()
            elif self.topo_visualization is False:
                self.topography_surface.VisibilityOn()
                self.topo_visualization = True
                self.interactor.Render()

        if key == 'q':
            print('closing vtk')
            self.close_window()

            # TODO: Decide if this even matter
            # # create render window, settings
            # self.renwin = vtk.vtkRenderWindow()
            # self.renwin.SetWindowName(self.ren_name)
            #
            # # Set 4 renderers. ie 3D, X,Y,Z projections
            # self.ren_list = self.create_ren_list()
            #
            # # create interactor and set interactor style, assign render window
            # self.interactor = vtk.vtkRenderWindowInteractor()
            # self.interactor.SetRenderWindow(self.renwin)
            # self.interactor.AddObserver("KeyPressEvent", self.key_callbacks)
            #
            # # 3d model camera for the 4 renders
            # self.camera_list = self._create_cameras(self.extent)
            # # Setting the camera and the background color to the renders
            # self.set_camera_backcolor()
            #
            # # Creating the axis
            # for e, r in enumerate(self.ren_list):
            #     # add axes actor to all renderer
            #     axe = self._create_axes(self.camera_list[e])
            #
            #     r.AddActor(axe)
            #     r.ResetCamera()

        if key == 'r':
            self.real_time = self.real_time ^ True

    def create_surface_points(self, vertices):
        """
        Method to create the points that form the surfaces
        Args:
            vertices (numpy.array): 2D array (XYZ) with the coordinates of the points

        Returns:
            vtk.vtkPoints: with the coordinates of the points
        """
        Points = vtk.vtkPoints()
        if self.ve != 1:
            vertices[:, 2] = vertices[:, 2] * self.ve
        #     raise NotImplementedError('Vertical exageration for surfaces not implemented yet.')
        # for v in vertices:
        #     v[-1] = self.ve * v[-1]
        #     Points.InsertNextPoint(v)
        Points.SetData(numpy_to_vtk(vertices))

        return Points

    @staticmethod
    def create_surface_triangles(simplices):
        """
        Method to create the Triangles that form the surfaces
        Args:
            simplices (numpy.array): 2D array with the value of the vertices that form every single triangle

        Returns:
            vtk.vtkTriangle
        """

        Triangles = vtk.vtkCellArray()
        Triangle = vtk.vtkTriangle()

        for s in simplices:
            Triangle.GetPointIds().SetId(0, s[0])
            Triangle.GetPointIds().SetId(1, s[1])
            Triangle.GetPointIds().SetId(2, s[2])

            Triangles.InsertNextCell(Triangle)
        return Triangles

    def create_surface(self, vertices, simplices, fn, alpha=.8):
        """
        Method to create the polydata that define the surfaces

        Args:
            vertices (numpy.array): 2D array (XYZ) with the coordinates of the points
            simplices (numpy.array): 2D array with the value of the vertices that form every single triangle
            fn (int): id
            alpha (float): Opacity

        Returns:
            vtk.vtkActor, vtk.vtkPolyDataMapper, vtk.vtkPolyData
        """
        vertices_c = copy.deepcopy(vertices)
        simplices_c = copy.deepcopy(simplices)

        surf_polydata = vtk.vtkPolyData()

        surf_polydata.SetPoints(self.create_surface_points(vertices_c))
        surf_polydata.SetPolys(self.create_surface_triangles(simplices_c))
        surf_polydata.Modified()

        surf_mapper = vtk.vtkPolyDataMapper()
        surf_mapper.SetInputData(surf_polydata)
        surf_mapper.Update()

        surf_actor = vtk.vtkActor()
        surf_actor.SetMapper(surf_mapper)
        surf_actor.GetProperty().SetColor(mcolors.hex2color(self.geo_model._surfaces.df.set_index('id')['color'][fn]))#self.C_LOT[fn])
        surf_actor.GetProperty().SetOpacity(alpha)

        return surf_actor, surf_mapper, surf_polydata

    def create_sphere(self, X, Y, Z, fn, n_sphere=0, n_render=0, n_index=0, r=0.03):
        """
        Method to create the sphere that represent the surface_points points
        Args:
            X: X coord
            Y: Y coord
            Z: Z corrd
            fn (int): id
            n_sphere (int): Number of the sphere
            n_render (int): Number of the render where the sphere belongs
            n_index (int): index value in the PandasDataframe of InupData.surface_points
            r (float): radius of the sphere

        Returns:
            vtk.vtkSphereWidget
        """
        s = vtk.vtkSphereWidget()
        s.SetInteractor(self.interactor)
        s.SetRepresentationToSurface()
        s.SetPriority(2)
        Z = Z * self.ve
        s.r_f = self._e_d_avrg * r
        s.PlaceWidget(X - s.r_f, X + s.r_f, Y - s.r_f, Y + s.r_f, Z - s.r_f, Z + s.r_f)
        s.GetSphereProperty().SetColor(mcolors.hex2color(self.geo_model._surfaces.df.set_index('id')['color'][fn]))#self.C_LOT[fn])

        s.SetCurrentRenderer(self.ren_list[n_render])
        s.n_sphere = n_sphere
        s.n_render = n_render
        s.index = n_index
        s.AddObserver("EndInteractionEvent", self.sphereCallback)  # EndInteractionEvent
        s.AddObserver("InteractionEvent", self.Callback_camera_reset)

        s.On()

        return s

    def create_foliation(self, X, Y, Z, fn,
                         Gx, Gy, Gz,
                         n_plane=0, n_render=0, n_index=0, alpha=0.5):
        """
        Method to create a plane given a foliation

        Args:
            X : X coord
            Y: Y coord
            Z: Z coord
            fn (int): id
            Gx (str): Component of the gradient x
            Gy (str): Component of the gradient y
            Gz (str): Component of the gradient z
            n_plane (int): Number of the plane
            n_render (int): Number of the render where the plane belongs
            n_index (int): index value in the PandasDataframe of InupData.surface_points
            alpha: Opacity of the plane

        Returns:
            vtk.vtkPlaneWidget
        """

        Z = Z * self.ve

        d = vtk.vtkPlaneWidget()
        d.SetInteractor(self.interactor)
        d.SetRepresentationToSurface()

        # Position
        source = vtk.vtkPlaneSource()

        source.SetNormal(Gx, Gy, Gz)
        source.SetCenter(X, Y, Z)
        a, b, c, d_, e, f = self.extent

        source.SetPoint1(X+self._e_dx*.01, Y-self._e_dy*.01, Z)
        source.SetPoint2(X-self._e_dx*.01, Y+self._e_dy*.01, Z)
        source.Update()
        d.SetInputData(source.GetOutput())
        d.SetHandleSize(.05)
        min_extent = np.min([self._e_dx, self._e_dy, self._e_dz])
        d.SetPlaceFactor(0.1)

        d.PlaceWidget(a, b, c, d_, e, f)
        d.SetNormal(Gx, Gy, Gz)
        d.SetCenter(X, Y, Z)
        d.GetPlaneProperty().SetColor(mcolors.hex2color(self.geo_model._surfaces.df.set_index('id')['color'][fn]))#self.C_LOT[fn])
        d.GetHandleProperty().SetColor(mcolors.hex2color(self.geo_model._surfaces.df.set_index('id')['color'][fn]))#self.C_LOT[fn])
        d.GetHandleProperty().SetOpacity(alpha)
        d.SetCurrentRenderer(self.ren_list[n_render])
        d.n_plane = n_plane
        d.n_render = n_render
        d.index = n_index
        d.AddObserver("EndInteractionEvent", self.planesCallback)
        d.AddObserver("InteractionEvent", self.Callback_camera_reset)

        d.On()

        return d

    def set_surfaces_old(self, vertices, simplices, alpha=1):
        """
        Create all the surfaces and set them to the corresponding renders for their posterior visualization with
        render_model

        Args:
            vertices (list): list of 3D numpy arrays containing the points that form each plane
            simplices (list): list of 3D numpy arrays containing the verticies that form every triangle
            surfaces (list): ordered list of strings containing the name of the surfaces to represent
            fns (list): ordered list of surface_numbers (int)
            alpha: Opacity of the plane

        Returns:
            None
        """
        self.surf_rend_1 = []

        surfaces = self.surface_name

        fns = self.geo_model._surface_points.df['id'].unique().squeeze()
        assert type(vertices) is list or type(vertices) is np.ndarray, 'vertices and simpleces have to be a list of' \
                                                                       ' arrays even when only one' \
                                                                       ' surface is passed'
        assert 'DefaultBasement' not in surfaces, 'Remove DefaultBasement from the list of surfaces'
        for v, s, fn in zip(vertices, simplices, np.atleast_1d(fns)):
            act, map, pol = self.create_surface(v, s, fn, alpha)
            self.surf_rend_1.append(act)

            self.ren_list[0].AddActor(act)
            self.ren_list[1].AddActor(act)
            self.ren_list[2].AddActor(act)
            self.ren_list[3].AddActor(act)

    def set_surfaces(self, surfaces, alpha=1):
        self.surf_rend_1 = []
        for idx, val in surfaces.df[['vertices', 'edges', 'id']].dropna().iterrows():
            act, map, pol = self.create_surface(val['vertices'], val['edges'], val['id'], alpha)
            self.surf_rend_1.append(act)

            self.ren_list[0].AddActor(act)
            self.ren_list[1].AddActor(act)
            self.ren_list[2].AddActor(act)
            self.ren_list[3].AddActor(act)

    def set_topography(self):
        # Create points on an XY grid with random Z coordinate
        vertices = copy.copy(self.geo_model._grid.topography.values)

        points = vtk.vtkPoints()
        # for v in vertices:
        #     v[-1] = v[-1]
        #     points.InsertNextPoint(v)
        if self.ve !=1:
            vertices[:, 2]= vertices[:, 2]*self.ve
        points.SetData(numpy_to_vtk(vertices))

        # Add the grid points to a polydata object
        polydata = vtk.vtkPolyData()
        polydata.SetPoints(points)
        #
        # glyphFilter = vtk.vtkVertexGlyphFilter()
        # glyphFilter.SetInputData(polydata)
        # glyphFilter.Update()
        #
        # # Create a mapper and actor
        # pointsMapper = vtk.vtkPolyDataMapper()
        # pointsMapper.SetInputConnection(glyphFilter.GetOutputPort())

        #
        # pointsActor = vtk.vtkActor()
        # pointsActor.SetMapper(pointsMapper)
        # pointsActor.GetProperty().SetPointSize(3)
        # pointsActor.GetProperty().SetColor(colors.GetColor3d("Red"))

        # Triangulate the grid points
        delaunay = vtk.vtkDelaunay2D()
        delaunay.SetInputData(polydata)
        delaunay.Update()

        # Create a mapper and actor
        triangulatedMapper = vtk.vtkPolyDataMapper()
        triangulatedMapper.SetInputConnection(delaunay.GetOutputPort())

        triangulatedActor = vtk.vtkActor()
        triangulatedActor.SetMapper(triangulatedMapper)

        self.topography_surface = triangulatedActor
        self._topography_polydata = polydata
        self._topography_delauny = delaunay
        self.ren_list[0].AddActor(triangulatedActor)
        self.ren_list[1].AddActor(triangulatedActor)
        self.ren_list[2].AddActor(triangulatedActor)
        self.ren_list[3].AddActor(triangulatedActor)
        try:
            if self.geo_model.solutions.geological_map is not None:
                self.set_geological_map()
        except AttributeError as ae:
            warnings.warn(str(ae))

    def set_geological_map(self):
        assert self.geo_model.solutions.geological_map is not None, 'Geological map not computed. First' \
                                                                    'set active the topography grid'
        arr_ = np.empty((0, 3), dtype='int')

        # Convert hex colors to rgb
        for idx, val in self.geo_model._surfaces.df['color'].iteritems():
            rgb = (255 * np.array(mcolors.hex2color(val)))
            arr_ = np.vstack((arr_, rgb))

        sel = np.round(self.geo_model.solutions.geological_map[0]).astype(int)[0]
        nv = numpy_to_vtk(arr_[sel - 1], array_type=3)
        self._topography_delauny.GetOutput().GetPointData().SetScalars(nv)

    def set_surface_points(self, indices=None):
        """
        Create all the surface_points points and set them to the corresponding renders for their posterior visualization
         with render_model

        Returns:
            None
        """

        if not indices:

            for e, val in enumerate(self.geo_model._surface_points.df.iterrows()):
                index = val[0]
                row = val[1]
                self.s_rend_1.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=0, n_index=index))
                self.s_rend_2.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=1, n_index=index))
                self.s_rend_3.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=2, n_index=index))
                self.s_rend_4.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=3, n_index=index))
        else:
            for e, val in enumerate(self.geo_model._surface_points.df.loc[np.atleast_1d(indices)].iterrows()):
                index = val[0]
                row = val[1]
                self.s_rend_1.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=0, n_index=index))
                self.s_rend_2.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=1, n_index=index))
                self.s_rend_3.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=2, n_index=index))
                self.s_rend_4.at[index] = (self.create_sphere(row['X'], row['Y'], row['Z'], row['id'],
                                           n_sphere=e, n_render=3, n_index=index))

    def set_orientations(self, indices=None):
        """
        Create all the orientations and set them to the corresponding renders for their posterior visualization with
        render_model
        Returns:
            None
        """

        if not indices:
            for e, val in enumerate(self.geo_model._orientations.df.iterrows()):
                index = val[0]
                row = val[1]
                self.o_rend_1.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=0, n_index=index))
                self.o_rend_2.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=1, n_index=index))
                self.o_rend_3.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=2, n_index=index))
                self.o_rend_4.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=3, n_index=index))
        else:
            for e, val in enumerate(self.geo_model._orientations.df.loc[np.atleast_1d(indices)].iterrows()):
                index = val[0]
                row = val[1]
                self.o_rend_1.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=0, n_index=index))
                self.o_rend_2.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=1, n_index=index))
                self.o_rend_3.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=2, n_index=index))
                self.o_rend_4.at[index] = (self.create_foliation(row['X'], row['Y'], row['Z'], row['id'],
                                           row['G_x'], row['G_y'], row['G_z'],
                                           n_plane=e, n_render=3, n_index=index))

    def create_slider_rep(self, min=0, max=10, val=0):

        slider_rep = vtk.vtkSliderRepresentation2D()
        slider_rep.SetMinimumValue(min)
        slider_rep.SetMaximumValue(max)
        slider_rep.SetValue(val)
        slider_rep.GetPoint1Coordinate().SetCoordinateSystemToDisplay()
        slider_rep.GetPoint1Coordinate().SetValue(0, 40)
        slider_rep.GetPoint2Coordinate().SetCoordinateSystemToDisplay()
        slider_rep.GetPoint2Coordinate().SetValue(300, 40)

        self.slider_w = vtk.vtkSliderWidget()
        self.slider_w.SetInteractor(self.interactor)
        self.slider_w.SetRepresentation(slider_rep)
        self.slider_w.SetCurrentRenderer(self.ren_list[0])
        self.slider_w.SetAnimationModeToAnimate()
        self.slider_w.On()
        self.slider_w.AddObserver('EndInteractionEvent', self.sliderCallback_traces)

    def create_slider_interactor(self, min=0, max=1, val=1):

        slider_rep = vtk.vtkSliderRepresentation2D()
        slider_rep.SetMinimumValue(min)
        slider_rep.SetMaximumValue(max)
        slider_rep.SetValue(val)
        slider_rep.GetPoint1Coordinate().SetCoordinateSystemToDisplay()
        slider_rep.GetPoint1Coordinate().SetValue(0, 60)
        slider_rep.GetPoint2Coordinate().SetCoordinateSystemToDisplay()
        slider_rep.GetPoint2Coordinate().SetValue(100, 60)
        slider_rep.SetTitleText('Interactor')

        self.slider_w = vtk.vtkSliderWidget()
        self.slider_w.SetInteractor(self.interactor)
        self.slider_w.SetRepresentation(slider_rep)
        self.slider_w.SetCurrentRenderer(self.ren_list[0])
        self.slider_w.SetAnimationModeToJump()

        self.slider_w.On()
        self.slider_w.AddObserver('EndInteractionEvent', self.sliderCallback_interactor)

    def sliderCallback_interactor(self, obj, event):
        if int(obj.GetRepresentation().GetValue()) == 0:
            self.interactor.ExitCallback()

    def sliderCallback_traces(self, obj, event):
        # TODO Check post class
        self.post.change_input_data(self.geo_model, obj.GetRepresentation().GetValue())
        try:
            # for surf in self.surf_rend_1:
            #     self.ren_list[0].RemoveActor(surf)
            #     self.ren_list[1].RemoveActor(surf)
            #     self.ren_list[2].RemoveActor(surf)
            #     self.ren_list[3].RemoveActor(surf)
            self.update_surfaces_real_time()
          #  ver, sim =
          #  self.set_surfaces(ver, sim)
        except AttributeError:
            print('no surf')
            pass
        try:
            for sph in zip(self.s_rend_1['val'], self.s_rend_2['val'], self.s_rend_3['val'],
                           self.s_rend_4['val'], self.geo_model._surface_points.df.iterrows()):

                row_i = sph[4][1]
                sph[0].PlaceWidget(row_i['X'] - sph[0].r_f, row_i['X'] + sph[0].r_f,
                                   row_i['Y'] - sph[0].r_f, row_i['Y'] + sph[0].r_f,
                                   row_i['Z'] - sph[0].r_f, row_i['Z'] + sph[0].r_f)

                sph[1].PlaceWidget(row_i['X'] - sph[1].r_f, row_i['X'] + sph[1].r_f,
                                   row_i['Y'] - sph[1].r_f, row_i['Y'] + sph[1].r_f,
                                   row_i['Z'] - sph[1].r_f, row_i['Z'] + sph[1].r_f)

                sph[2].PlaceWidget(row_i['X'] - sph[2].r_f, row_i['X'] + sph[2].r_f,
                                   row_i['Y'] - sph[2].r_f, row_i['Y'] + sph[2].r_f,
                                   row_i['Z'] - sph[2].r_f, row_i['Z'] + sph[2].r_f)

                sph[3].PlaceWidget(row_i['X'] - sph[3].r_f, row_i['X'] + sph[3].r_f,
                                   row_i['Y'] - sph[3].r_f, row_i['Y'] + sph[3].r_f,
                                   row_i['Z'] - sph[3].r_f, row_i['Z'] + sph[3].r_f)
        except AttributeError:
            pass
        try:
            for fol in (zip(self.f_rend_1, self.f_rend_2, self.f_rend_3, self.f_rend_4, self.geo_model._orientations.iterrows())):
                row_f = fol[4][1]

                fol[0].SetCenter(row_f['X'], row_f['Y'], row_f['Z'])
                fol[0].SetNormal(row_f['G_x'], row_f['G_y'], row_f['G_z'])

        except AttributeError:
            pass

    def sphereCallback(self, obj, event):
        """
        Function that rules what happens when we move a sphere. At the moment we update the other 3 renderers and
        update the pandas data frame.
        """
        #self.interactor.ExitCallback()

       # self.Callback_camera_reset()

        # Get new position of the sphere
        new_center = obj.GetCenter()

        # Get which sphere we are moving
        index = obj.index

        # Check what render we are working with
        render = obj.n_render

        # This must be the radius
        #r_f = obj.r_f

        self.SphereCallback_change_df(index, new_center)
        self.SphereCallbak_move_changes(index)

        if self.real_time:
            try:
                self.update_surfaces_real_time()
                #vertices, simpleces =
                #self.set_surfaces(vertices, simpleces)
            except AssertionError:
                print('Not enough data to compute the model')

    def Callback_camera_reset(self,  obj, event):

        # Resetting the xy camera when a sphere is moving to be able to change only 2D
        fp = self.ren_list[1].GetActiveCamera().GetFocalPoint()
        p = self.ren_list[1].GetActiveCamera().GetPosition()
        dist = np.sqrt((p[0] - fp[0]) ** 2 + (p[1] - fp[1]) ** 2 + (p[2] - fp[2]) ** 2)
        self.ren_list[1].GetActiveCamera().SetPosition(fp[0], fp[1], fp[2] + dist)
        self.ren_list[1].GetActiveCamera().SetViewUp(0.0, 1.0, 0.0)

        # Resetting the yz camera when a sphere is moving to be able to change only 2D
        fp = self.ren_list[2].GetActiveCamera().GetFocalPoint()
        p = self.ren_list[2].GetActiveCamera().GetPosition()
        dist = np.sqrt((p[0] - fp[0]) ** 2 + (p[1] - fp[1]) ** 2 + (p[2] - fp[2]) ** 2)
        self.ren_list[2].GetActiveCamera().SetPosition(fp[0] + dist, fp[1], fp[2])
        self.ren_list[2].GetActiveCamera().SetViewUp(0.0, -1.0, 0.0)
        self.ren_list[2].GetActiveCamera().Roll(90)

        # Resetting the xz camera when a sphere is moving to be able to change only 2D
        fp = self.ren_list[3].GetActiveCamera().GetFocalPoint()
        p = self.ren_list[3].GetActiveCamera().GetPosition()
        dist = np.sqrt((p[0] - fp[0]) ** 2 + (p[1] - fp[1]) ** 2 + (p[2] - fp[2]) ** 2)
        self.ren_list[3].GetActiveCamera().SetPosition(fp[0], fp[1] - dist, fp[2])
        self.ren_list[3].GetActiveCamera().SetViewUp(-1.0, 0.0, 0.0)
        self.ren_list[3].GetActiveCamera().Roll(90)

    def SphereCallback_change_df(self, index, new_center):
        index = np.atleast_1d(index)
        # Modify Pandas DataFrame
        self.geo_model.modify_surface_points(index, X=[new_center[0]], Y=[new_center[1]], Z=[new_center[2]])

    def SphereCallbak_move_changes(self, indices):
       # print(indices)
        df_changes = self.geo_model._surface_points.df.loc[np.atleast_1d(indices)][['X', 'Y', 'Z', 'id']]
        for index, df_row in df_changes.iterrows():
            new_center = df_row[['X', 'Y', 'Z']].values

            # Update renderers
            s1 = self.s_rend_1.loc[index, 'val']

            s1.PlaceWidget(new_center[0] - s1.r_f, new_center[0] + s1.r_f,
                           new_center[1] - s1.r_f, new_center[1] + s1.r_f,
                           new_center[2] - s1.r_f, new_center[2] + s1.r_f)

            s1.GetSphereProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][df_row['id']]))#self.C_LOT[df_row['id']])

            s2 = self.s_rend_2.loc[index, 'val']
            s2.PlaceWidget(new_center[0] - s2.r_f, new_center[0] + s2.r_f,
                           new_center[1] - s2.r_f, new_center[1] + s2.r_f,
                           new_center[2] - s2.r_f, new_center[2] + s2.r_f)

            s2.GetSphereProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][df_row['id']]))

            s3 = self.s_rend_3.loc[index, 'val']
            s3.PlaceWidget(new_center[0] - s3.r_f, new_center[0] + s3.r_f,
                           new_center[1] - s3.r_f, new_center[1] + s3.r_f,
                           new_center[2] - s3.r_f, new_center[2] + s3.r_f)

            s3.GetSphereProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][df_row['id']]))

            s4 = self.s_rend_4.loc[index, 'val']
            s4.PlaceWidget(new_center[0] - s4.r_f, new_center[0] + s4.r_f,
                           new_center[1] - s4.r_f, new_center[1] + s4.r_f,
                           new_center[2] - s4.r_f, new_center[2] + s4.r_f)

            s4.GetSphereProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][df_row['id']]))

    def SphereCallback_delete_point(self, ind_i):
        """
        Warning this deletion system will lead to memory leaks until the vtk object is reseted (hopefully). This is
        mainly a vtk problem to delete objects
        """
        ind_i = np.atleast_1d(ind_i)
        # Deactivating widget
        for i in ind_i:
            self.s_rend_1.loc[i, 'val'].Off()
            self.s_rend_2.loc[i, 'val'].Off()
            self.s_rend_3.loc[i, 'val'].Off()
            self.s_rend_4.loc[i, 'val'].Off()

        self.s_rend_1.drop(ind_i)
        self.s_rend_2.drop(ind_i)
        self.s_rend_3.drop(ind_i)
        self.s_rend_4.drop(ind_i)

    def planesCallback(self, obj, event):
        """
        Function that rules what happens when we move a plane. At the moment we update the other 3 renderers and
        update the pandas data frame
        """

      # self.Callback_camera_reset()

        # Get new position of the plane and GxGyGz
        new_center = obj.GetCenter()
        new_normal = obj.GetNormal()
        # Get which plane we are moving
        index = obj.index

        self.planesCallback_change_df(index, new_center, new_normal)
        self.planesCallback_move_changes(index)

        if self.real_time:
            # try:
            #     if self.real_time:
            #         for surf in self.surf_rend_1:
            #             self.ren_list[0].RemoveActor(surf)
            #             self.ren_list[1].RemoveActor(surf)
            #             self.ren_list[2].RemoveActor(surf)
            #             self.ren_list[3].RemoveActor(surf)
            # except AttributeError:
            #     pass

            try:
                self.update_surfaces_real_time()
              #  vertices, simpleces =
              #  self.set_surfaces(vertices, simpleces)
            except AssertionError:
                print('Not enough data to compute the model')

    def planesCallback_change_df(self, index, new_center, new_normal):

        # Modify Pandas DataFrame
        # update the gradient vector components and its location
        self.geo_model.modify_orientations(index, X=new_center[0], Y=new_center[1], Z=new_center[2],
                                           G_x=new_normal[0], G_y=new_normal[1], G_z=new_normal[2])
        # update the dip and azimuth values according to the new gradient
        # self.geo_model.calculate_orientations()

    def planesCallback_move_changes(self, indices):
        df_changes = self.geo_model._orientations.df.loc[np.atleast_1d(indices)][['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'id']]
        for index, new_values_df in df_changes.iterrows():
            new_center = new_values_df[['X', 'Y', 'Z']].values
            new_normal = new_values_df[['G_x', 'G_y', 'G_z']].values
            new_source = vtk.vtkPlaneSource()
            new_source.SetCenter(new_center)
            new_source.SetNormal(new_normal)
            new_source.Update()

            plane1 = self.o_rend_1.loc[index, 'val']
          #  plane1.SetInputData(new_source.GetOutput())
            plane1.SetNormal(new_normal)
            plane1.SetCenter(new_center[0], new_center[1], new_center[2])
            plane1.GetPlaneProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))#self.C_LOT[new_values_df['id']])
            plane1.GetHandleProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))

            plane2 = self.o_rend_2.loc[index, 'val']
            plane2.SetInputData(new_source.GetOutput())
            plane2.SetNormal(new_normal)
            plane2.SetCenter(new_center[0], new_center[1], new_center[2])
            plane2.GetPlaneProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))
            plane2.GetHandleProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))

            plane3 = self.o_rend_3.loc[index, 'val']
            plane3.SetInputData(new_source.GetOutput())
            plane3.SetNormal(new_normal)
            plane3.SetCenter(new_center[0], new_center[1], new_center[2])
            plane3.GetPlaneProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))
            plane3.GetHandleProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))

            plane4 = self.o_rend_4.loc[index, 'val']
            plane4.SetInputData(new_source.GetOutput())
            plane4.SetNormal(new_normal)
            plane4.SetCenter(new_center[0], new_center[1], new_center[2])
            plane4.GetPlaneProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))
            plane4.GetHandleProperty().SetColor(mcolors.hex2color(
                self.geo_model._surfaces.df.set_index('id')['color'][new_values_df['id']]))

    def planesCallback_delete_point(self, ind_o):
        """
        Warning this deletion system will lead to memory leaks until the vtk object is reseted (hopefully). This is
        mainly a vtk problem to delete objects
        """
        ind_o = np.atleast_1d(ind_o)
        # Deactivating widget
        for i in ind_o:
            self.o_rend_1.loc[i, 'val'].Off()
            self.o_rend_2.loc[i, 'val'].Off()
            self.o_rend_3.loc[i, 'val'].Off()
            self.o_rend_4.loc[i, 'val'].Off()

        self.s_rend_1.drop(ind_o)
        self.s_rend_2.drop(ind_o)
        self.s_rend_3.drop(ind_o)
        self.s_rend_4.drop(ind_o)

    def create_ren_list(self):
        """
        Create a list of the 4 renderers we use. One general view and 3 cartersian projections
        Returns:
            list: list of renderers
        """

        # viewport dimensions setup
        xmins = [0, 0.6, 0.6, 0.6]
        xmaxs = [0.6, 1, 1, 1]
        ymins = [0, 0, 0.33, 0.66]
        ymaxs = [1, 0.33, 0.66, 1]

        # create list of renderers, set vieport values
        ren_list = []
        for i in range(self.n_ren):
            # append each renderer to list of renderers
            ren_list.append(vtk.vtkRenderer())
            # add each renderer to window
            self.renwin.AddRenderer(ren_list[-1])
            # set viewport for each renderer
            ren_list[-1].SetViewport(xmins[i], ymins[i], xmaxs[i], ymaxs[i])

        return ren_list

    def _create_cameras(self, extent, verbose=0):
        """
        Create the 4 cameras for each renderer
        """
        _e = extent
        _e_dx = _e[1] - _e[0]
        _e_dy = _e[3] - _e[2]
        _e_dz = _e[5] - _e[4]
        _e_d_avrg = (_e_dx + _e_dy + _e_dz) / 3
        _e_max = np.argmax(_e)

        # General camera
        model_cam = vtk.vtkCamera()
        model_cam.SetPosition(_e[_e_max] * 5, _e[_e_max] * 5, _e[_e_max] * 5)
        model_cam.SetFocalPoint(np.min(_e[0:2]) + _e_dx / 2,
                                np.min(_e[2:4]) + _e_dy / 2,
                                np.min(_e[4:]) + _e_dz / 2)

        model_cam.SetViewUp(-0.239, 0.155, 0.958)

        # XY camera RED
        xy_cam = vtk.vtkCamera()

        xy_cam.SetPosition(np.min(_e[0:2]) + _e_dx / 2,
                           np.min(_e[2:4]) + _e_dy / 2,
                           _e[_e_max] * 4)

        xy_cam.SetFocalPoint(np.min(_e[0:2]) + _e_dx / 2,
                             np.min(_e[2:4]) + _e_dy / 2,
                             np.min(_e[4:]) + _e_dz / 2)

        # YZ camera GREEN
        yz_cam = vtk.vtkCamera()
        yz_cam.SetPosition(_e[_e_max] * 4,
                           np.min(_e[2:4]) + _e_dy / 2,
                           np.min(_e[4:]) + _e_dz / 2)

        yz_cam.SetFocalPoint(np.min(_e[0:2]) + _e_dx / 2,
                             np.min(_e[2:4]) + _e_dy / 2,
                             np.min(_e[4:]) + _e_dz / 2)
        yz_cam.SetViewUp(0, -1, 0)
        yz_cam.Roll(90)

        # XZ camera BLUE
        xz_cam = vtk.vtkCamera()
        xz_cam.SetPosition(np.min(_e[0:2]) + _e_dx / 2,
                           - _e[_e_max] * 4,
                           np.min(_e[4:]) + _e_dz / 2)

        xz_cam.SetFocalPoint(np.min(_e[0:2]) + _e_dx / 2,
                             np.min(_e[2:4]) + _e_dy / 2,
                             np.min(_e[4:]) + _e_dz / 2)
        xz_cam.SetViewUp(0, 1, 0)
        xz_cam.Roll(0)

        # camera position debugging
        if verbose == 1:
            print("RED XY:", xy_cam.GetPosition())
            print("RED FP:", xy_cam.GetFocalPoint())
            print("GREEN YZ:", yz_cam.GetPosition())
            print("GREEN FP:", yz_cam.GetFocalPoint())
            print("BLUE XZ:", xz_cam.GetPosition())
            print("BLUE FP:", xz_cam.GetFocalPoint())

        return [model_cam, xy_cam, yz_cam, xz_cam]

    def set_camera_backcolor(self, color=None):
        """
        define background colors of the renderers
        """
        if color is None:
            color = (66 / 250, 66 / 250, 66 / 250)

        ren_color = [color for i in range(self.n_ren)]

        for i in range(self.n_ren):
            # set active camera for each renderer
            self.ren_list[i].SetActiveCamera(self.camera_list[i])
            # set background color for each renderer
            self.ren_list[i].SetBackground(ren_color[i][0], ren_color[i][1], ren_color[i][2])

    def _create_axes(self, camera, verbose=0, tick_vis=False):
        """
        Create the axes boxes
        """
        cube_axes_actor = vtk.vtkCubeAxesActor()
        cube_axes_actor.SetBounds(self.extent)
        cube_axes_actor.SetCamera(camera)
        if verbose == 1:
            print(cube_axes_actor.GetAxisOrigin())

        # set axes and label colors
        cube_axes_actor.GetTitleTextProperty(0).SetColor(1.0, 0.0, 0.0)
        cube_axes_actor.GetLabelTextProperty(0).SetColor(1.0, 0.0, 0.0)

        cube_axes_actor.GetTitleTextProperty(1).SetColor(0.0, 1.0, 0.0)
        cube_axes_actor.GetLabelTextProperty(1).SetColor(0.0, 1.0, 0.0)
        cube_axes_actor.GetTitleTextProperty(2).SetColor(0.0, 0.0, 1.0)
        cube_axes_actor.GetLabelTextProperty(2).SetColor(0.0, 0.0, 1.0)

        cube_axes_actor.DrawXGridlinesOn()
        cube_axes_actor.DrawYGridlinesOn()
        cube_axes_actor.DrawZGridlinesOn()

        if not tick_vis:
            cube_axes_actor.XAxisMinorTickVisibilityOff()
            cube_axes_actor.YAxisMinorTickVisibilityOff()
            cube_axes_actor.ZAxisMinorTickVisibilityOff()

        cube_axes_actor.SetXTitle("X")
        cube_axes_actor.SetYTitle("Y")
        cube_axes_actor.SetZTitle("Z")

        cube_axes_actor.SetXAxisLabelVisibility(1)
        cube_axes_actor.SetYAxisLabelVisibility(1)
        cube_axes_actor.SetZAxisLabelVisibility(1)

        # only plot grid lines furthest from viewpoint
        # ensure platform compatibility for the grid line options
        if sys.platform == "win32":
            cube_axes_actor.SetGridLineLocation(cube_axes_actor.VTK_GRID_LINES_FURTHEST)
        else:  # rather use elif == "linux" ? but what about other platforms
            try:  # apparently this can also go wrong on linux, maybe depends on vtk version?
                cube_axes_actor.SetGridLineLocation(cube_axes_actor.VTK_GRID_LINES_FURTHEST)
            except AttributeError:
                pass

        return cube_axes_actor

    def delete_surfaces(self):
        try:
            for surf in self.surf_rend_1:
                self.ren_list[0].RemoveActor(surf)
                self.ren_list[1].RemoveActor(surf)
                self.ren_list[2].RemoveActor(surf)
                self.ren_list[3].RemoveActor(surf)
        except AttributeError:
            pass

    def update_surfaces_real_time(self, delete=True):

        if delete is True:
            self.delete_surfaces()
        try:
            gp.compute_model(self.geo_model, sort_surfaces=False, compute_mesh=True)
        except IndexError:
            print('IndexError: Model not computed. Laking data in some surface')
        except AssertionError:
            print('AssertionError: Model not computed. Laking data in some surface')
        # try:
        #     v_l, s_l = self.geo_model.surfaces.df['vertices'], self.geo_model.surfaces.df['edges']
        # except IndexError:
        #     try:
        #         v_l, s_l = self.geo_model.surfaces.df['vertices'], self.geo_model.surfaces.df['edges']
        #     except IndexError:
        #         v_l, s_l = self.geo_model.surfaces.df['vertices'], self.geo_model.surfaces.df['edges']

        self.set_surfaces(self.geo_model._surfaces)
        # if self.geo_model.solutions.geological_map is not None:
        #     try:
        #         self.set_geological_map()
        #     except AttributeError:
        #         pass
        return True

    @staticmethod
    def export_vtk_lith_block(geo_data, lith_block=None, path="./default"):
        """
        Export data to a vtk file for posterior visualizations

        Args:
            geo_data(:class:''): All values of a DataManagement object
            block(numpy.array): 3D array containing the lithology block
            path (str): path to the location of the vtk

        Returns:
            None
        """



        # Dimensions

        nx, ny, nz = geo_data._grid.regular_grid.resolution

        lx = geo_data._grid.regular_grid.extent[1] - geo_data._grid.regular_grid.extent[0]
        ly = geo_data._grid.regular_grid.extent[3] - geo_data._grid.regular_grid.extent[2]
        lz = geo_data._grid.regular_grid.extent[5] - geo_data._grid.regular_grid.extent[4]

        dx, dy, dz = lx / nx, ly / ny, lz / nz

        ncells = nx * ny * nz

        npoints = (nx + 1) * (ny + 1) * (nz + 1)

        # Coordinates
        x = np.arange(geo_data._grid.regular_grid.extent[0], geo_data._grid.regular_grid.extent[1] + 0.1, dx, dtype='float64')

        y = np.arange(geo_data._grid.regular_grid.extent[2], geo_data._grid.regular_grid.extent[3] + 0.1, dy, dtype='float64')

        z = np.arange(geo_data._grid.regular_grid.extent[4], geo_data._grid.regular_grid.extent[5] + 0.1, dz, dtype='float64')

        if lith_block is None:
            lith_block = geo_data.solutions.lith_block
        lith = lith_block.reshape((nx, ny, nz)).flatten(order='F')
        out = pv.RectilinearGrid(x,y,z)
        out['Lithology'] = lith
        out.save(path+'_lith_block.vtr')

    @staticmethod
    def export_vtk_surfaces(geo_model, vertices=None, simplices=None, path=None, name='_surfaces', alpha=1):
        """
        Export data to a vtk file for posterior visualizations

        Args:
            geo_model(gempy.InputData): All values of a DataManagement object
            block(numpy.array): 3D array containing the lithology block
            path (str): path to the location of the vtk

        Returns:
            None
        """
        import vtk

        if vertices is None or simplices is None:
            vertices = geo_model.solutions.vertices
            simplices = geo_model.solutions.edges

        s_n = 0
        for e, values in enumerate(vertices):
            # setup points and vertices
            s_n += 1
            Points = vtk.vtkPoints()
            Triangles = vtk.vtkCellArray()
            Triangle = vtk.vtkTriangle()
            for p in values:
                Points.InsertNextPoint(p)

            # Unfortunately in this simple example the following lines are ambiguous.
            # The first 0 is the index of the triangle vertex which is ALWAYS 0-2.
            # The second 0 is the index into the point (geometry) array, so this can range from 0-(NumPoints-1)
            # i.e. a more general statement is triangle->GetPointIds()->SetId(0, PointId);
            for i in simplices[e]:
                Triangle.GetPointIds().SetId(0, i[0])
                Triangle.GetPointIds().SetId(1, i[1])
                Triangle.GetPointIds().SetId(2, i[2])

                Triangles.InsertNextCell(Triangle)

            polydata = vtk.vtkPolyData()
            polydata.SetPoints(Points)
            polydata.SetPolys(Triangles)

            polydata.Modified()
            if vtk.VTK_MAJOR_VERSION <= 5:
                polydata.Update()

            writer = vtk.vtkXMLPolyDataWriter();

            # Add colors
            surf_mapper = vtk.vtkPolyDataMapper()
            surf_mapper.SetInputData(polydata)
            surf_mapper.Update()

            surf_actor = vtk.vtkActor()
            surf_actor.SetMapper(surf_mapper)
            surf_actor.GetProperty().SetColor(mcolors.hex2color(geo_model._surfaces.df.set_index('id')['color'][s_n]))
            surf_actor.GetProperty().SetOpacity(alpha)

            if not path:
                path = "./default_"

            writer.SetFileName(path+'_surfaces_'+str(e)+'.vtp')
            if vtk.VTK_MAJOR_VERSION <= 5:
                writer.SetInput(polydata)
            else:
                writer.SetInputData(polydata)
            writer.Write()


class GemPyvtkInteract(vtkVisualization):

    def resume(self):
        self.interactor.Start()

    def restart(self, render_surfaces=True, **kwargs):
        self.set_surface_points()
        self.set_orientations()
        if render_surfaces is True:
            self.set_surfaces(self.geo_model._surfaces)

        self.render_model(**kwargs)

    def set_real_time_on(self):

        self.real_time = True

    def set_real_time_off(self):
        self.real_time = False

    def render_move_surface_points(self, indices):
        self.SphereCallbak_move_changes(indices)
       # print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_add_surface_points(self, indices):
        self.set_surface_points(indices)
       # print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_delete_surface_points(self, indices):
        self.SphereCallback_delete_point(indices)
      #  print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_move_orientations(self, indices):
        self.planesCallback_move_changes(indices)
      #  print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_add_orientations(self, indices):
        self.set_orientations(indices)
     #   print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_delete_orientations(self, indices):
        self.planesCallback_delete_point(indices)
     #   print('vtk-gempy real time is:' + str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.interactor.Render()

    def render_surfaces(self, alpha=1):
        self.delete_surfaces()
        self.set_surfaces(self.geo_model._surfaces, alpha=alpha)
        self.interactor.Render()

    def render_topography(self):
        try:
            self.ren_list[0].RemoveActor(self.topography_surface)
            self.ren_list[1].RemoveActor(self.topography_surface)
            self.ren_list[2].RemoveActor(self.topography_surface)
            self.ren_list[3].RemoveActor(self.topography_surface)
        except AttributeError:
            pass

      #  print('vtk-gempy real time is:' +str(self.real_time))
        if self.real_time is True:
            self.update_surfaces_real_time()
        self.set_topography()
        self.interactor.Render()

    def update_model(self):
        if self.real_time is True:
            self.update_surfaces_real_time()
            self.interactor.Render()


class steno3D():
    def __init__(self, geo_data, project, **kwargs ):
        if STENO_IMPORT is False:
            raise ImportError( 'Steno 3D package is not installed. No 3D online visualization available.')
        description = kwargs.get('description', 'Nothing')

        self._data = geo_data
        self.surfaces = pn.DataFrame.from_dict(geo_data.get_surface_number(), orient='index')


        steno3d.login()

        self.proj = steno3d.Project(
            title=project,
            description=description,
            public=True,
        )

    def plot3D_steno_grid(self, block, plot=False, **kwargs):


        mesh = steno3d.Mesh3DGrid(h1=np.ones(self._data.resolution[0]) * (self._data.extent[1] - self._data.extent[0]) /
                                                                         (self._data.resolution[0] - 1),
                                  h2=np.ones(self._data.resolution[1]) * (self._data.extent[3] - self._data.extent[2]) /
                                                                         (self._data.resolution[1] - 1),
                                  h3=np.ones(self._data.resolution[2]) * (self._data.extent[5] - self._data.extent[4]) /
                                                                         (self._data.resolution[2] - 1),
                                  O=[self._data.extent[0], self._data.extent[2], self._data.extent[4]])

        data = steno3d.DataArray(
            title='Lithologies_block',
            array=block)

        vol = steno3d.Volume(project=self.proj, mesh=mesh, data=[dict(location='CC', data=data)])
       # vol.upload()

        if plot:
            return vol.plot()

    def plot3D_steno_surface(self, ver, sim):

        for surface in self.surfaces.iterrows():
            if surface[1].values[0] == 0:
                pass

            #for vertices, simpleces in zip(ver[surface[1].values[0]], sim[surface[1].values[0]]):
            surface_mesh = steno3d.Mesh2D(
                vertices=ver[surface[1].values[0]-1],
                triangles=sim[surface[1].values[0]-1],
                opts=dict(wireframe=False)
            )
            surface_obj = steno3d.Surface(
                project=self.proj,
                title='Surface: {}'.format(surface[0]),
                mesh=surface_mesh,
                opts=dict(opacity=1)
            )


def get_fault_ellipse_params(fault_points:np.ndarray):
    """Get the fault ellipse parameters a and b from griven fault points (should
    be the rotated ones.

    Args:
        fault_points (np.ndarray): Fault points

    Returns:
        (tuple) main axis scalars of fault ellipse a,b
    """
    a = (fault_points[:, 0].max() - fault_points[:, 0].min()) / 2
    b = (fault_points[:, 1].max() - fault_points[:, 1].min()) / 2
    return a, b


def get_fault_rotation_objects(geo_model, fault:str):
    """Gets fault rotation objects: rotation matrix U, the rotated fault points,
    rotated centroid, and the ellipse parameters a and b.

    Args:
        geo_model (gempy.core.model.Project): gempy geo_model object
        fault (str): Name of the fault surface.

    Returns:
        U (np.ndarray): Rotation matrix.
        rfpts (np.ndarray): Rotated fault points.
        rctr (np.array): Centroid of the rotated fault points.
        a (float): Horizontal ellipse parameter.
        b (float): Vertical ellipse parameter.
    """
    filter_ = geo_model._surface_points.df.surface == fault
    fpts = geo_model._surface_points.df[filter_][["X", "Y", "Z"]].values.T
    ctr = np.mean(fpts, axis=1)
    x = fpts - ctr.reshape((-1, 1))
    M = np.dot(x, x.T)
    U = np.linalg.svd(M)
    rfpts = np.dot(fpts.T, U[0])
    # rfpts = np.dot(rfpts, U[-1])
    rctr = np.mean(rfpts, axis=0)

    a, b = get_fault_ellipse_params(rfpts)
    return U, rfpts, rctr, a, b


def cut_finite_fault_surfaces(geo_model, ver:dict, sim:dict):
    """Cut vertices and simplices for finite fault surfaces to finite fault ellipse

    Args:
        geo_model (gempy.core.model.Project): gempy geo_model object
        ver (dict): Dictionary with surfaces as keys and vertices ndarray as values.
        sim (dict): Dictionary with surfaces as keys and simplices ndarray as values.

    Returns:
        ver, sim (dict, dict): Updated vertices and simplices with finite fault
            surfaces cut to ellipses.
    """
    from scipy.spatial import Delaunay
    from copy import copy

    finite_ver = copy(ver)
    finite_sim = copy(sim)

    finite_fault_series = list(geo_model._faults.df[geo_model._faults.df["isFinite"] == True].index)
    finite_fault_surfaces = list(
        geo_model._surfaces.df[geo_model._surfaces.df._stack == finite_fault_series].surface.unique())

    for fault in finite_fault_surfaces:
        U, fpoints_rot, fctr_rot, a, b = get_fault_rotation_objects(geo_model, "Fault 1")
        rpoints = np.dot(ver[fault], U[0])
        # rpoints = np.dot(rpoints, U[-1])
        r = (rpoints[:, 0] - fctr_rot[0]) ** 2 / a ** 2 + (rpoints[:, 1] - fctr_rot[1]) ** 2 / b ** 2

        finite_ver[fault] = finite_ver[fault][r < 1]
        delaunay = Delaunay(finite_ver[fault])
        finite_sim[fault] = delaunay.simplices
        # finite_sim[fault] = finite_sim[fault][np.isin(sim[fault], np.argwhere(r<0.33))]

    return finite_ver, finite_sim
